@@ -1,6 +1,6 @@
 /*****************************************************************************/
 /*  rfc2068.c - General purpose routines for the HTTP protocol               */
-/*  Copyright (C) 1998-2003 Brian Masney <masneyb@gftp.org>                  */
+/*  Copyright (C) 1998-2007 Brian Masney <masneyb@gftp.org>                  */
 /*                                                                           */
 /*  This program is free software; you can redistribute it and/or modify     */
 /*  it under the terms of the GNU General Public License as published by     */
@@ -20,7 +20,7 @@
 #include "gftp.h"
 #include "httpcommon.h"
 
-static const char cvsid[] = "$Id: rfc2068.c,v 1.41 2004/03/17 11:59:27 masneyb Exp $";
+static const char cvsid[] = "$Id: rfc2068.c 950 2008-01-24 23:28:30Z masneyb $";
 
 static gftp_config_vars config_vars[] =
 {
@@ -53,15 +53,12 @@ static int
 rfc2068_connect (gftp_request * request)
 {
   char *proxy_hostname, *proxy_config;
-  rfc2068_params * params;
   intptr_t proxy_port;
   int ret;
 
   g_return_val_if_fail (request != NULL, GFTP_EFATAL);
   g_return_val_if_fail (request->hostname != NULL, GFTP_EFATAL);
 
-  params = request->protocol_data;
-
   if (request->datafd > 0)
     return (0);
 
@@ -96,8 +93,9 @@ static off_t
 rfc2068_read_response (gftp_request * request)
 {
   rfc2068_params * params;
+  unsigned int chunked;
   char tempstr[8192];
-  int ret, chunked;
+  int ret;
 
   params = request->protocol_data;
   *tempstr = '\0';
@@ -147,7 +145,7 @@ rfc2068_read_response (gftp_request * request)
                                 sizeof (tempstr), request->datafd)) < 0)
         return (ret);
 
-      if (sscanf ((char *) tempstr, "%lx", &params->chunk_size) != 1)
+      if (sscanf (tempstr, GFTP_OFF_T_HEX_PRINTF_MOD, &params->chunk_size) != 1)
         {
           request->logging_function (gftp_logging_recv, request,
                                      _("Received wrong response from server, disconnecting\nInvalid chunk size '%s' returned by the remote server\n"), 
@@ -164,7 +162,7 @@ rfc2068_read_response (gftp_request * request)
         {
           params->extra_read_buffer_len = params->chunk_size * -1;
           params->chunk_size = 0;
-          params->extra_read_buffer = g_malloc (params->extra_read_buffer_len + 1);
+          params->extra_read_buffer = g_malloc0 ((gulong) params->extra_read_buffer_len + 1);
           memcpy (params->extra_read_buffer, params->rbuf->curpos + (params->rbuf->cur_bufsize - params->extra_read_buffer_len), params->extra_read_buffer_len);
           params->extra_read_buffer[params->extra_read_buffer_len] = '\0';
           params->rbuf->cur_bufsize -= params->extra_read_buffer_len;
@@ -178,7 +176,7 @@ rfc2068_read_response (gftp_request * request)
 
 
 static off_t 
-rfc2068_send_command (gftp_request * request, const void *command, size_t len)
+rfc2068_send_command (gftp_request * request, const char *command)
 {
   char *tempstr, *str, *proxy_hostname, *proxy_username, *proxy_password;
   intptr_t proxy_port;
@@ -268,13 +266,14 @@ rfc2068_disconnect (gftp_request * request)
 
 
 static off_t
-rfc2068_get_file (gftp_request * request, const char *filename, int fd,
+rfc2068_get_file (gftp_request * request, const char *filename,
                   off_t startsize)
 {
   char *tempstr, *oldstr, *hf;
-  int restarted;
-  intptr_t use_http11;
   rfc2068_params * params;
+  intptr_t use_http11;
+  int restarted;
+  size_t len;
   off_t size;
 
   g_return_val_if_fail (request != NULL, GFTP_EFATAL);
@@ -284,10 +283,7 @@ rfc2068_get_file (gftp_request * request, const char *filename, int fd,
 
   gftp_lookup_request_option (request, "use_http11", &use_http11);
 
-  if (fd > 0)
-    request->datafd = fd;
-
-  hf = gftp_build_path (request->hostname, filename, NULL);
+  hf = gftp_build_path (request, request->hostname, filename, NULL);
 
   if (request->username == NULL || *request->username == '\0')
     tempstr = g_strconcat ("GET ", request->url_prefix, "://", hf,
@@ -301,36 +297,26 @@ rfc2068_get_file (gftp_request * request, const char *filename, int fd,
 
   if (use_http11 && startsize > 0)
     {
-#if defined (_LARGEFILE_SOURCE)
       request->logging_function (gftp_logging_misc, request,
-                              _("Starting the file transfer at offset %lld\n"),
-                              (long long) startsize);
+                              _("Starting the file transfer at offset " GFTP_OFF_T_PRINTF_MOD "\n"),
+                              startsize);
 
       oldstr = tempstr;
-      tempstr = g_strdup_printf ("%sRange: bytes=%lld-\n", tempstr, startsize);
+      tempstr = g_strdup_printf ("%sRange: bytes=" GFTP_OFF_T_PRINTF_MOD "-\n",
+                                 tempstr, startsize);
       g_free (oldstr);
-#else
-      request->logging_function (gftp_logging_misc, request,
-			       _("Starting the file transfer at offset %ld\n"), 
-                               startsize);
-
-      oldstr = tempstr;
-      tempstr = g_strdup_printf ("%sRange: bytes=%ld-\n", tempstr, startsize);
-      g_free (oldstr);
-#endif
     }
 
-  size = rfc2068_send_command (request, tempstr, strlen (tempstr));
+  size = rfc2068_send_command (request, tempstr);
   g_free (tempstr);
   if (size < 0)
     return (size);
 
   restarted = 0;
-  if (strlen (request->last_ftp_response) > 9 
-      && strncmp (request->last_ftp_response + 9, "206", 3) == 0)
+  len = strlen (request->last_ftp_response);
+  if (len  > 9 && strncmp (request->last_ftp_response + 9, "206", 3) == 0)
     restarted = 1;
-  else if (strlen (request->last_ftp_response) < 9 ||
-           strncmp (request->last_ftp_response + 9, "200", 3) != 0)
+  else if (len < 9 || strncmp (request->last_ftp_response + 9, "200", 3) != 0)
     {
       request->logging_function (gftp_logging_error, request,
 			         _("Cannot retrieve file %s\n"), filename);
@@ -410,7 +396,7 @@ rfc2068_list_files (gftp_request * request)
   if (strncmp (request->directory, "/", strlen (request->directory)) == 0)
     hd = g_strdup (request->hostname);
   else
-    hd = gftp_build_path (request->hostname, request->directory, NULL);
+    hd = gftp_build_path (request, request->hostname, request->directory, NULL);
 
   if (request->username == NULL || *request->username == '\0')
     tempstr = g_strconcat ("GET ", request->url_prefix, "://", hd,
@@ -422,7 +408,7 @@ rfc2068_list_files (gftp_request * request)
 
   g_free (hd);
 
-  ret = rfc2068_send_command (request, tempstr, strlen (tempstr));
+  ret = rfc2068_send_command (request, tempstr);
   g_free (tempstr);
   if (ret < 0)
     return ((int) ret);
@@ -445,7 +431,6 @@ rfc2068_list_files (gftp_request * request)
 static off_t 
 rfc2068_get_file_size (gftp_request * request, const char *filename)
 {
-  rfc2068_params *params;
   char *tempstr, *hf;
   intptr_t use_http11;
   off_t size;
@@ -453,10 +438,9 @@ rfc2068_get_file_size (gftp_request * request, const char *filename)
   g_return_val_if_fail (request != NULL, GFTP_EFATAL);
   g_return_val_if_fail (filename != NULL, GFTP_EFATAL);
 
-  params = request->protocol_data;
   gftp_lookup_request_option (request, "use_http11", &use_http11);
 
-  hf = gftp_build_path (request->hostname, filename, NULL);
+  hf = gftp_build_path (request, request->hostname, filename, NULL);
 
   if (request->username == NULL || *request->username == '\0')
     tempstr = g_strconcat ("HEAD ", request->url_prefix, "://", hf,
@@ -468,7 +452,7 @@ rfc2068_get_file_size (gftp_request * request, const char *filename)
 
   g_free (hf);
 
-  size = rfc2068_send_command (request, tempstr, strlen (tempstr));
+  size = rfc2068_send_command (request, tempstr);
   g_free (tempstr);
   return (size);
 }
@@ -502,12 +486,11 @@ parse_html_line (char *tempstr, gftp_file * fle)
      is a directory or not */
   if (*(pos - 1) == '/')
     {
-      fle->attribs = g_strdup ("drwxr-xr-x");
+      fle->st_mode = S_IFDIR | S_IRUSR | S_IWUSR | S_IXUSR | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH;
       *(pos - 1) = '\0';
-      fle->isdir = 1;
     }
   else
-    fle->attribs = g_strdup ("-rw-r--r--");
+    fle->st_mode = S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH;
 
   /* Copy filename */
   if (strchr (stpos, '/') != NULL || strncmp (stpos, "mailto:", 7) == 0 ||
@@ -607,7 +590,7 @@ rfc2068_get_next_file (gftp_request * request, gftp_file * fle, int fd)
 {
   rfc2068_params * params;
   char tempstr[8192];
-  ssize_t len;
+  size_t len;
   int ret;
 
   g_return_val_if_fail (request != NULL, GFTP_EFATAL);
@@ -629,14 +612,14 @@ rfc2068_get_next_file (gftp_request * request, gftp_file * fle, int fd)
         return (ret);
 
       if (parse_html_line (tempstr, fle) == 0 || fle->file == NULL)
-	gftp_file_destroy (fle);
+	gftp_file_destroy (fle, 0);
       else
 	break;
     }
 
   if (fle->file == NULL)
     {
-      gftp_file_destroy (fle);
+      gftp_file_destroy (fle, 0);
       return (0);
     }
 
@@ -654,27 +637,26 @@ rfc2068_get_next_file (gftp_request * request, gftp_file * fle, int fd)
 static int
 rfc2068_chdir (gftp_request * request, const char *directory)
 {
-  char *tempstr, *olddir;
+  char *tempstr;
 
   g_return_val_if_fail (request != NULL, GFTP_EFATAL);
   g_return_val_if_fail (directory != NULL, GFTP_EFATAL);
 
-  if (request->directory != directory)
+  if (*directory != '/' && request->directory != NULL)
     {
-      olddir = request->directory;
-
-      if (*directory != '/' && request->directory != NULL)
-        {
-          tempstr = g_strconcat (request->directory, "/", directory, NULL);
-          request->directory = expand_path (tempstr);
-          g_free (tempstr);
-        }
-      else
-        request->directory = expand_path (directory);
-
-      if (olddir != NULL)
-        g_free (olddir);
+      tempstr = g_strconcat (request->directory, "/", directory, NULL);
+      g_free (request->directory);
+      request->directory = gftp_expand_path (request, tempstr);
+      g_free (tempstr);
     }
+  else
+    {
+      if (request->directory != NULL)
+        g_free (request->directory);
+
+      request->directory = gftp_expand_path (request, directory);
+    }
+
   return (0);
 }
 
@@ -719,6 +701,7 @@ rfc2068_chunked_read (gftp_request * request, void *ptr, size_t size, int fd)
   char *stpos, *crlfpos;
   void *read_ptr_pos;
   ssize_t retval;
+  size_t sret;
 
   params = request->protocol_data;
   params->read_ref_cnt++;
@@ -778,9 +761,8 @@ rfc2068_chunked_read (gftp_request * request, void *ptr, size_t size, int fd)
           params->read_ref_cnt--;
           return (retval);
         }
-      else if (retval > 0)
-        params->read_bytes += retval;
 
+      params->read_bytes += retval;
       if (params->chunk_size > 0)
         {
           params->chunk_size -= retval;
@@ -788,23 +770,23 @@ rfc2068_chunked_read (gftp_request * request, void *ptr, size_t size, int fd)
           return (retval);
         }
 
-      retval += begin_ptr_len;
+      sret = retval + begin_ptr_len;
     }
   else
-    retval = begin_ptr_len;
+    sret = begin_ptr_len;
 
-  ((char *) ptr)[retval] = '\0';
+  ((char *) ptr)[sret] = '\0';
 
   if (!params->chunked_transfer)
     {
       params->read_ref_cnt--;
-      return (retval);
+      return (sret);
     }
 
   stpos = (char *) ptr;
   while (params->chunk_size == 0)
     {
-      current_size = retval - (stpos - (char *) ptr);
+      current_size = sret - (stpos - (char *) ptr);
       if (current_size == 0)
         break;
 
@@ -818,13 +800,13 @@ rfc2068_chunked_read (gftp_request * request, void *ptr, size_t size, int fd)
           /* The current chunk size is split between multiple packets.
              Save this chunk and read the next */
 
-          params->extra_read_buffer = g_malloc (current_size + 1);
+          params->extra_read_buffer = g_malloc0 ((gulong) current_size + 1);
           memcpy (params->extra_read_buffer, stpos, current_size);
           params->extra_read_buffer[current_size] = '\0';
           params->extra_read_buffer_len = current_size;
-          retval -= current_size;
+          sret -= current_size;
 
-          if (retval == 0)
+          if (sret == 0)
             {
               /* Don't let a hostile web server send us in an infinite recursive
                  loop */
@@ -838,16 +820,21 @@ rfc2068_chunked_read (gftp_request * request, void *ptr, size_t size, int fd)
                 }
 
               retval = rfc2068_chunked_read (request, ptr, size, fd);
+              if (retval < 0)
+                return (retval);
+
+              sret = retval;
             }
 
           params->read_ref_cnt--;
-          return (retval);
+          return (sret);
         }
 
       *crlfpos = '\0';
       crlfpos++; /* advance to line feed */
 
-      if (sscanf (stpos + 2, "%lx", &params->chunk_size) != 1)
+      if (sscanf (stpos + 2, GFTP_OFF_T_HEX_PRINTF_MOD,
+                  &params->chunk_size) != 1)
         {
           request->logging_function (gftp_logging_recv, request,
                                      _("Received wrong response from server, disconnecting\nInvalid chunk size '%s' returned by the remote server\n"), 
@@ -857,8 +844,8 @@ rfc2068_chunked_read (gftp_request * request, void *ptr, size_t size, int fd)
           return (GFTP_EFATAL);
         }
 
-      crlfsize = crlfpos - (char *) stpos + 1;
-      retval -= crlfsize;
+      crlfsize = crlfpos - stpos + 1;
+      sret -= crlfsize;
       current_size -= crlfsize;
 
       if (params->chunk_size == 0)
@@ -871,7 +858,7 @@ rfc2068_chunked_read (gftp_request * request, void *ptr, size_t size, int fd)
 
           params->eof = 1;
           params->read_ref_cnt--;
-          return (retval);
+          return (sret);
         }
 
       memmove (stpos, crlfpos + 1, current_size + 1);
@@ -886,7 +873,7 @@ rfc2068_chunked_read (gftp_request * request, void *ptr, size_t size, int fd)
     }
 
   params->read_ref_cnt--;
-  return (retval);
+  return (sret);
 }
 
 
@@ -920,8 +907,10 @@ rfc2068_init (gftp_request * request)
   request->put_next_file_chunk = NULL;
   request->end_transfer = rfc2068_end_transfer;
   request->abort_transfer = rfc2068_end_transfer; /* NOTE: uses end_transfer */
+  request->stat_filename = NULL;
   request->list_files = rfc2068_list_files;
   request->get_next_file = rfc2068_get_next_file;
+  request->get_next_dirlist_line = NULL;
   request->get_file_size = rfc2068_get_file_size;
   request->chdir = rfc2068_chdir;
   request->rmdir = NULL;
@@ -935,9 +924,11 @@ rfc2068_init (gftp_request * request)
   request->set_config_options = rfc2068_set_config_options;
   request->url_prefix = g_strdup ("http");
   request->need_hostport = 1;
-  request->need_userpass = 0;
+  request->need_username = 0;
+  request->need_password = 0;
   request->use_cache = 1;
   request->always_connected = 1;
+  request->use_local_encoding = 0;
 
   request->protocol_data = g_malloc0 (sizeof (rfc2068_params));
   params = request->protocol_data;
