@@ -1,6 +1,6 @@
 /*****************************************************************************/
 /*  sshv2.c - functions that will use the sshv2 protocol                     */
-/*  Copyright (C) 1998-2003 Brian Masney <masneyb@gftp.org>                  */
+/*  Copyright (C) 1998-2007 Brian Masney <masneyb@gftp.org>                  */
 /*                                                                           */
 /*  This program is free software; you can redistribute it and/or modify     */
 /*  it under the terms of the GNU General Public License as published by     */
@@ -18,7 +18,7 @@
 /*****************************************************************************/
 
 #include "gftp.h"
-static const char cvsid[] = "$Id: sshv2.c,v 1.70 2004/12/31 22:55:23 masneyb Exp $";
+static const char cvsid[] = "$Id: sshv2.c 950 2008-01-24 23:28:30Z masneyb $";
 
 #define SSH_MAX_HANDLE_SIZE		256
 #define SSH_MAX_STRING_SIZE		34000
@@ -140,14 +140,14 @@ typedef struct sshv2_params_tag
 
 
 static char *
-sshv2_initialize_string (gftp_request * request, size_t len)
+sshv2_initialize_buffer (gftp_request * request, size_t len)
 {
   sshv2_params * params;
   guint32 num;
   char *ret;
 
   params = request->protocol_data;
-  ret = g_malloc0 (len + 1);
+  ret = g_malloc0 ((gulong) len + 1);
 
   num = htonl (params->id++);
   memcpy (ret, &num, 4);
@@ -157,13 +157,91 @@ sshv2_initialize_string (gftp_request * request, size_t len)
 
 
 static void
-sshv2_add_string_to_buf (char *buf, const char *str)
+sshv2_add_string_to_buf (char *buf, const char *str, size_t len)
 {
   guint32 num;
 
   num = htonl (strlen (str));
   memcpy (buf, &num, 4);
-  strcpy (buf + 4, str);
+  memcpy (buf + 4, str, len);
+}
+
+
+static char *
+sshv2_initialize_buffer_with_i18n_string (gftp_request * request,
+                                          const char *str, size_t *msglen)
+{
+  const char *addstr;
+  char *utf8, *ret;
+  size_t pathlen;
+
+  utf8 = gftp_filename_from_utf8 (request, str, &pathlen);
+  if (utf8 != NULL)
+    addstr = utf8;
+  else
+    {
+      addstr = str;
+      pathlen = strlen (str);
+    }
+
+  *msglen += pathlen + 8;
+  ret = sshv2_initialize_buffer (request, *msglen);
+  sshv2_add_string_to_buf (ret + 4, addstr, pathlen);
+
+  if (utf8 != NULL)
+    g_free (utf8);
+
+  return (ret);
+}
+
+
+static char *
+_sshv2_generate_utf8_path (gftp_request * request, const char *str, size_t *len)
+{
+  char *path, *utf8;
+
+  if (*str == '/')
+    path = g_strdup (str);
+  else
+    path = gftp_build_path (request, request->directory, str, NULL);
+
+  utf8 = gftp_filename_from_utf8 (request, path, len);
+  if (utf8 != NULL)
+    {
+      g_free (path);
+      return (utf8);
+    }
+  else
+    {
+      *len = strlen (path);
+      return (path);
+    }
+}
+
+
+static char *
+sshv2_initialize_buffer_with_two_i18n_paths (gftp_request * request,
+                                             const char *str1,
+                                             const char *str2, size_t *msglen)
+{
+  char *first_utf8, *sec_utf8, *ret;
+  size_t pathlen1, pathlen2;
+
+  first_utf8 = _sshv2_generate_utf8_path (request, str1, &pathlen1);
+  sec_utf8 = _sshv2_generate_utf8_path (request, str2, &pathlen2);
+
+  *msglen += pathlen1 + pathlen2 + 12;
+  ret = sshv2_initialize_buffer (request, *msglen);
+  sshv2_add_string_to_buf (ret + 4, first_utf8, pathlen1);
+  sshv2_add_string_to_buf (ret + 8 + pathlen1, sec_utf8, pathlen2);
+
+  if (first_utf8 != str1)
+    g_free (first_utf8);
+
+  if (sec_utf8 != str2)
+    g_free (sec_utf8);
+
+  return (ret);
 }
 
 
@@ -172,27 +250,20 @@ sshv2_initialize_string_with_path (gftp_request * request, const char *path,
                                    size_t *len, char **endpos)
 {
   char *ret, *tempstr;
-  size_t pathlen;
+  size_t origlen;
 
+  origlen = *len;
   if (*path == '/')
-    {
-      pathlen = strlen (path);
-      *len += pathlen + 8;
-      ret = sshv2_initialize_string (request, *len);
-      sshv2_add_string_to_buf (ret + 4, path);
-    }
+    ret = sshv2_initialize_buffer_with_i18n_string (request, path, len);
   else
     {
       tempstr = gftp_build_path (request, request->directory, path, NULL);
-      pathlen = strlen (tempstr);
-      *len += pathlen + 8;
-      ret = sshv2_initialize_string (request, *len);
-      sshv2_add_string_to_buf (ret + 4, tempstr);
+      ret = sshv2_initialize_buffer_with_i18n_string (request, tempstr, len);
       g_free (tempstr);
     }
 
   if (endpos != NULL)
-    *endpos = ret + 8 + pathlen;
+    *endpos = ret + *len - origlen;
 
   return (ret);
 }
@@ -211,7 +282,7 @@ sshv2_add_exec_args (char **logstr, size_t *logstr_len, char ***args,
   va_end (argp);
 
   *logstr_len += strlen (tempstr);
-  *logstr = g_realloc (*logstr, *logstr_len + 1);
+  *logstr = g_realloc (*logstr, (gulong) *logstr_len + 1);
   strcat (*logstr, tempstr);
 
   curpos = tempstr;
@@ -270,7 +341,7 @@ sshv2_gen_exec_args (gftp_request * request)
 
   args_len = 1;
   args_cur = 15;
-  args = g_malloc (sizeof (char *) * args_cur);
+  args = g_malloc0 (sizeof (char *) * args_cur);
   args[0] = g_strdup (tempstr);
 
   logstr = g_strdup (args[0]);
@@ -317,7 +388,7 @@ sshv2_start_login_sequence (gftp_request * request, int fdm, int ptymfd)
 
   rem = len = SSH_LOGIN_BUFSIZE;
   diff = 0;
-  tempstr = g_malloc0 (len + 1);
+  tempstr = g_malloc0 ((gulong) len + 1);
   wrotepw = 0;
   ok = 1;
 
@@ -346,16 +417,26 @@ sshv2_start_login_sequence (gftp_request * request, int fdm, int ptymfd)
       FD_SET (ptymfd, &eset);
 
       ret = select (maxfd + 1, &rset, NULL, &eset, NULL);
-      if (ret < 0 && (errno == EINTR || errno == EAGAIN))
-        continue;
-
       if (ret < 0)
         {
-          request->logging_function (gftp_logging_error, request,
-                                     _("Connection to %s timed out\n"),
-                                     request->hostname);
-          gftp_disconnect (request);
-          return (GFTP_ERETRYABLE);
+          if (errno == EINTR || errno == EAGAIN)
+            {
+              if (request->cancel)
+                {
+                  gftp_disconnect (request);
+                  return (GFTP_ERETRYABLE);
+                }
+
+              continue;
+            }
+          else
+            {
+              request->logging_function (gftp_logging_error, request,
+                                         _("Connection to %s timed out\n"),
+                                         request->hostname);
+              gftp_disconnect (request);
+              return (GFTP_ERETRYABLE);
+            }
         }
 
       if (FD_ISSET (fdm, &eset) || FD_ISSET (ptymfd, &eset))
@@ -394,6 +475,7 @@ sshv2_start_login_sequence (gftp_request * request, int fdm, int ptymfd)
             break;
         }
 
+      clear_tempstr = 0;
       if (pwstrs[pwidx] != NULL)
         {
           clear_tempstr = 1;
@@ -464,8 +546,6 @@ sshv2_start_login_sequence (gftp_request * request, int fdm, int ptymfd)
           tempstr = g_realloc (tempstr, len);
           continue;
         }
-      else
-        clear_tempstr = 0;
 
       if (clear_tempstr)
         {
@@ -532,6 +612,7 @@ sshv2_log_command (gftp_request * request, gftp_logging_level level,
       case SSH_FXP_CLOSE:
         request->logging_function (level, request, 
                                    _("%d: Close\n"), id);
+        break;
       case SSH_FXP_OPENDIR:
         request->logging_function (level, request, 
                                    _("%d: Open Directory %s\n"), id,
@@ -756,7 +837,7 @@ sshv2_read_response (gftp_request * request, sshv2_message * message,
     }
 
   message->command = buf[4];
-  message->buffer = g_malloc (message->length + 1);
+  message->buffer = g_malloc0 (message->length + 1);
 
   message->pos = message->buffer;
   message->end = message->buffer + message->length - 1;
@@ -825,6 +906,7 @@ static int
 sshv2_read_status_response (gftp_request * request, sshv2_message * message,
                             int fd, int fxp_is_retryable, int fxp_is_wrong)
 {
+  guint32 num;
   int ret;
 
   memset (message, 0, sizeof (*message));
@@ -833,8 +915,15 @@ sshv2_read_status_response (gftp_request * request, sshv2_message * message,
     return (ret);
   else if (fxp_is_retryable > 0 && ret == fxp_is_retryable)
     {
+      memcpy (&num, message->buffer + 4, 4);
+      num = ntohl (num);
+
       sshv2_message_free (message);
-      return (GFTP_ERETRYABLE);
+
+      if (num == SSH_FX_PERMISSION_DENIED)
+        return (GFTP_EFATAL);
+      else
+        return (GFTP_ERETRYABLE);
     }
   else if (fxp_is_wrong > 0 && ret != fxp_is_wrong)
     return (sshv2_wrong_response (request, message));
@@ -852,10 +941,11 @@ sshv2_response_return_code (gftp_request * request, sshv2_message * message,
       case SSH_FX_OK:
       case SSH_FX_EOF:
       case SSH_FX_NO_SUCH_FILE:
-      case SSH_FX_PERMISSION_DENIED:
       case SSH_FX_FAILURE:
       case SSH_FX_OP_UNSUPPORTED:
         return (GFTP_ERETRYABLE);
+      case SSH_FX_PERMISSION_DENIED:
+        return (GFTP_EFATAL);
       default:
         return (sshv2_wrong_response (request, message));
     }
@@ -942,7 +1032,7 @@ sshv2_buffer_get_string (gftp_request * request, sshv2_message * message,
 
   if (return_string)
     {
-      string = g_malloc (len + 1);
+      string = g_malloc0 (len + 1);
       memcpy (string, message->pos, len);
       string[len] = '\0';
     }
@@ -955,10 +1045,36 @@ sshv2_buffer_get_string (gftp_request * request, sshv2_message * message,
 
 
 static int
-sshv2_getcwd (gftp_request * request)
+sshv2_send_command_and_check_response (gftp_request * request, char type,
+                                       char *command, size_t len)
 {
   sshv2_message message;
-  char *tempstr, *dir;
+  int ret;
+
+  ret = sshv2_send_command (request, type, command, len);
+  if (ret < 0)
+    return (ret);
+
+  memset (&message, 0, sizeof (message));
+  if ((ret = sshv2_read_response (request, &message, -1)) < 0)
+    return (ret);
+
+  message.pos += 4;
+  if ((ret = sshv2_buffer_get_int32 (request, &message, SSH_FX_OK, 1,
+                                     NULL)) < 0)
+    return (ret);
+
+  sshv2_message_free (&message);
+
+  return (0);
+}
+
+
+static int
+sshv2_getcwd (gftp_request * request)
+{
+  char *tempstr, *dir, *utf8;
+  sshv2_message message;
   size_t len;
   int ret;
 
@@ -970,12 +1086,9 @@ sshv2_getcwd (gftp_request * request)
   else
     dir = request->directory;
 
-  len = strlen (dir) + 8;
-  tempstr = sshv2_initialize_string (request, len);
-  sshv2_add_string_to_buf (tempstr + 4, dir);
-
+  len = 0;
+  tempstr = sshv2_initialize_buffer_with_i18n_string (request, dir, &len);
   ret = sshv2_send_command (request, SSH_FXP_REALPATH, tempstr, len);
-
   g_free (tempstr);
   if (ret < 0)
     return (ret);
@@ -995,9 +1108,18 @@ sshv2_getcwd (gftp_request * request)
   if ((ret = sshv2_buffer_get_int32 (request, &message, 1, 1, NULL)) < 0)
     return (ret);
 
-  if ((request->directory = sshv2_buffer_get_string (request, &message, 1)) == NULL)
+  if ((dir = sshv2_buffer_get_string (request, &message, 1)) == NULL)
     return (GFTP_EFATAL);
 
+  utf8 = gftp_filename_to_utf8 (request, dir, &len);
+  if (utf8 != NULL)
+    {
+      request->directory = utf8;
+      g_free (dir);
+    }
+  else
+    request->directory = dir;
+
   sshv2_message_free (&message);
   return (0);
 }
@@ -1192,7 +1314,7 @@ sshv2_list_files (gftp_request * request)
   sshv2_params * params;
   sshv2_message message;
   char *tempstr;
-  size_t len;
+  size_t msglen;
   int ret;
 
   g_return_val_if_fail (request != NULL, GFTP_EFATAL);
@@ -1204,12 +1326,11 @@ sshv2_list_files (gftp_request * request)
   request->logging_function (gftp_logging_misc, request,
 			     _("Retrieving directory listing...\n"));
 
-  len = strlen (request->directory) + 8;
-  tempstr = sshv2_initialize_string (request, len);
-  sshv2_add_string_to_buf (tempstr + 4, request->directory);
-
-  ret = sshv2_send_command (request, SSH_FXP_OPENDIR, tempstr, len);
-
+  msglen = 0;
+  tempstr = sshv2_initialize_buffer_with_i18n_string (request,
+                                                      request->directory,
+                                                      &msglen);
+  ret = sshv2_send_command (request, SSH_FXP_OPENDIR, tempstr, msglen);
   g_free (tempstr);
   if (ret < 0)
     return (ret);
@@ -1356,7 +1477,7 @@ sshv2_get_next_file (gftp_request * request, gftp_file * fle, int fd)
 
       if (!request->cached)
         {
-          request->last_dir_entry = g_malloc (params->message.length + 4);
+          request->last_dir_entry = g_malloc0 (params->message.length + 4);
           len = htonl (params->message.length);
           memcpy (request->last_dir_entry, &len, 4);
           request->last_dir_entry[4] = params->message.command;
@@ -1416,38 +1537,32 @@ sshv2_chdir (gftp_request * request, const char *directory)
   g_return_val_if_fail (request != NULL, GFTP_EFATAL);
   g_return_val_if_fail (request->protonum == GFTP_SSHV2_NUM, GFTP_EFATAL);
 
-  if (request->directory != directory)
-    {
-      len = 0;
-      tempstr = sshv2_initialize_string_with_path (request, directory,
-                                                   &len, NULL);
+  len = 0;
+  tempstr = sshv2_initialize_string_with_path (request, directory, &len, NULL);
 
-      ret = sshv2_send_command (request, SSH_FXP_REALPATH, tempstr, len);
+  ret = sshv2_send_command (request, SSH_FXP_REALPATH, tempstr, len);
 
-      g_free (tempstr);
-      if (ret < 0)
-        return (ret);
+  g_free (tempstr);
+  if (ret < 0)
+    return (ret);
 
-      ret = sshv2_read_status_response (request, &message, -1, SSH_FXP_STATUS,
-                                        SSH_FXP_NAME);
-      if (ret < 0)
-        return (ret);
+  ret = sshv2_read_status_response (request, &message, -1, SSH_FXP_STATUS,
+                                    SSH_FXP_NAME);
+  if (ret < 0)
+    return (ret);
 
-      message.pos += 4;
-      if ((ret = sshv2_buffer_get_int32 (request, &message, 1, 1, NULL)) < 0)
-        return (ret);
+  message.pos += 4;
+  if ((ret = sshv2_buffer_get_int32 (request, &message, 1, 1, NULL)) < 0)
+    return (ret);
 
-      if ((dir = sshv2_buffer_get_string (request, &message, 1)) == NULL)
-        return (GFTP_EFATAL);
+  if ((dir = sshv2_buffer_get_string (request, &message, 1)) == NULL)
+    return (GFTP_EFATAL);
 
-      if (request->directory)
-        g_free (request->directory);
-
-      request->directory = dir;
-      sshv2_message_free (&message);
-      return (0);
-    }
+  if (request->directory)
+    g_free (request->directory);
 
+  request->directory = dir;
+  sshv2_message_free (&message);
   return (0);
 }
 
@@ -1455,7 +1570,6 @@ sshv2_chdir (gftp_request * request, const char *directory)
 static int 
 sshv2_rmdir (gftp_request * request, const char *directory)
 {
-  sshv2_message message;
   char *tempstr;
   size_t len;
   int ret;
@@ -1467,30 +1581,16 @@ sshv2_rmdir (gftp_request * request, const char *directory)
   len = 0;
   tempstr = sshv2_initialize_string_with_path (request, directory, &len, NULL);
 
-  ret = sshv2_send_command (request, SSH_FXP_RMDIR, tempstr, len);
-
+  ret = sshv2_send_command_and_check_response (request, SSH_FXP_RMDIR,
+                                               tempstr, len);
   g_free (tempstr);
-  if (ret < 0)
-    return (ret);
-
-  memset (&message, 0, sizeof (message));
-  if ((ret = sshv2_read_response (request, &message, -1)) < 0)
-    return (ret);
-
-  message.pos += 4;
-  if ((ret = sshv2_buffer_get_int32 (request, &message, SSH_FX_OK, 1, NULL)) < 0)
-    return (ret);
-
-  sshv2_message_free (&message);
-
-  return (0);
+  return (ret);
 }
 
 
 static int 
 sshv2_rmfile (gftp_request * request, const char *file)
 {
-  sshv2_message message;
   char *tempstr;
   size_t len;
   int ret;
@@ -1502,23 +1602,10 @@ sshv2_rmfile (gftp_request * request, const char *file)
   len = 0;
   tempstr = sshv2_initialize_string_with_path (request, file, &len, NULL);
 
-  ret = sshv2_send_command (request, SSH_FXP_REMOVE, tempstr, len);
-
+  ret = sshv2_send_command_and_check_response (request, SSH_FXP_REMOVE,
+                                               tempstr, len);
   g_free (tempstr);
-  if (ret < 0)
-    return (ret);
-
-  memset (&message, 0, sizeof (message));
-  if ((ret = sshv2_read_response (request, &message, -1)) < 0)
-    return (ret);
-
-  message.pos += 4;
-  if ((ret = sshv2_buffer_get_int32 (request, &message, SSH_FX_OK, 1, NULL)) < 0)
-    return (ret);
-
-  sshv2_message_free (&message);
-
-  return (0);
+  return (ret);
 }
 
 
@@ -1526,7 +1613,6 @@ static int
 sshv2_chmod (gftp_request * request, const char *file, mode_t mode)
 {
   char *tempstr, *endpos;
-  sshv2_message message;
   guint32 num;
   size_t len;
   int ret;
@@ -1544,30 +1630,16 @@ sshv2_chmod (gftp_request * request, const char *file, mode_t mode)
   num = htonl (mode);
   memcpy (endpos + 4, &num, 4);
 
-  ret = sshv2_send_command (request, SSH_FXP_SETSTAT, tempstr, len);
-
+  ret = sshv2_send_command_and_check_response (request, SSH_FXP_SETSTAT,
+                                               tempstr, len);
   g_free (tempstr);
-  if (ret < 0)
-    return (ret);
-
-  memset (&message, 0, sizeof (message));
-  if ((ret = sshv2_read_response (request, &message, -1)) < 0)
-    return (ret);
-
-  message.pos += 4;
-  if ((ret = sshv2_buffer_get_int32 (request, &message, SSH_FX_OK, 1, NULL)) < 0)
-    return (ret);
-
-  sshv2_message_free (&message);
-
-  return (0);
+  return (ret);
 }
 
 
 static int 
 sshv2_mkdir (gftp_request * request, const char *newdir)
 {
-  sshv2_message message;
   char *tempstr;
   size_t len;
   int ret;
@@ -1582,32 +1654,18 @@ sshv2_mkdir (gftp_request * request, const char *newdir)
   /* No need to set attributes since all characters of the tempstr buffer is
      initialized to 0 */
 
-  ret = sshv2_send_command (request, SSH_FXP_MKDIR, tempstr, len);
-
+  ret = sshv2_send_command_and_check_response (request, SSH_FXP_MKDIR,
+                                               tempstr, len);
   g_free (tempstr);
-  if (ret < 0)
-    return (ret);
-
-  memset (&message, 0, sizeof (message));
-  if ((ret = sshv2_read_response (request, &message, -1)) < 0)
-    return (ret);
-
-  message.pos += 4;
-  if ((ret = sshv2_buffer_get_int32 (request, &message, SSH_FX_OK, 1, NULL)) < 0)
-    return (ret);
-
-  sshv2_message_free (&message);
-
-  return (0);
+  return (ret);
 }
 
 
 static int 
 sshv2_rename (gftp_request * request, const char *oldname, const char *newname)
 {
-  char *tempstr, *oldstr, *newstr;
-  size_t oldlen, newlen, len;
-  sshv2_message message;
+  char *tempstr;
+  size_t msglen;
   int ret;
 
   g_return_val_if_fail (request != NULL, GFTP_EFATAL);
@@ -1615,44 +1673,15 @@ sshv2_rename (gftp_request * request, const char *oldname, const char *newname)
   g_return_val_if_fail (oldname != NULL, GFTP_EFATAL);
   g_return_val_if_fail (newname != NULL, GFTP_EFATAL);
 
-  if (*oldname == '/')
-    oldstr = g_strdup (oldname);
-  else
-    oldstr = gftp_build_path (request, request->directory, oldname, NULL);
-
-  if (*newname == '/')
-    newstr = g_strdup (newname);
-  else
-    newstr = gftp_build_path (request, request->directory, newname, NULL);
-
-  oldlen = strlen (oldstr); 
-  newlen = strlen (newstr); 
-
-  len = oldlen + newlen + 12;
-  tempstr = sshv2_initialize_string (request, len);
-  sshv2_add_string_to_buf (tempstr + 4, oldstr);
-  sshv2_add_string_to_buf (tempstr + 8 + oldlen, newstr);
-
-  g_free (oldstr);
-  g_free (newstr);
-
-  ret = sshv2_send_command (request, SSH_FXP_RENAME, tempstr, len);
+  msglen = 0;
+  tempstr = sshv2_initialize_buffer_with_two_i18n_paths (request,
+                                                         oldname, newname,
+                                                         &msglen);
 
+  ret = sshv2_send_command_and_check_response (request, SSH_FXP_RENAME,
+                                               tempstr, msglen);
   g_free (tempstr);
-  if (ret < 0)
-    return (ret);
-
-  memset (&message, 0, sizeof (message));
-  if ((ret = sshv2_read_response (request, &message, -1)) < 0)
-    return (ret);
-
-  message.pos += 4;
-  if ((ret = sshv2_buffer_get_int32 (request, &message, SSH_FX_OK, 1, NULL)) < 0)
-    return (ret);
-
-  sshv2_message_free (&message);
-
-  return (0);
+  return (ret);
 }
 
 
@@ -1660,7 +1689,6 @@ static int
 sshv2_set_file_time (gftp_request * request, const char *file, time_t datetime)
 {
   char *tempstr, *endpos;
-  sshv2_message message;
   guint32 num;
   size_t len;
   int ret;
@@ -1679,23 +1707,10 @@ sshv2_set_file_time (gftp_request * request, const char *file, time_t datetime)
   memcpy (endpos + 4, &num, 4);
   memcpy (endpos + 8, &num, 4);
 
-  ret = sshv2_send_command (request, SSH_FXP_SETSTAT, tempstr, len);
-
+  ret = sshv2_send_command_and_check_response (request, SSH_FXP_SETSTAT,
+                                               tempstr, len);
   g_free (tempstr);
-  if (ret < 0)
-    return (ret);
-
-  memset (&message, 0, sizeof (message));
-  if ((ret = sshv2_read_response (request, &message, -1)) < 0)
-    return (ret);
-
-  message.pos += 4;
-  if ((ret = sshv2_buffer_get_int32 (request, &message, SSH_FX_OK, 1, NULL)) < 0)
-    return (ret);
-
-  sshv2_message_free (&message);
-
-  return (0);
+  return (ret);
 }
 
 
@@ -1744,7 +1759,7 @@ sshv2_send_stat_command (gftp_request * request, const char *filename,
 
 static int
 sshv2_stat_filename (gftp_request * request, const char *filename,
-                     mode_t * mode)
+                     mode_t * mode, off_t * filesize)
 {
   gftp_file fle;
   int ret;
@@ -1755,6 +1770,8 @@ sshv2_stat_filename (gftp_request * request, const char *filename,
     return (ret);
 
   *mode = fle.st_mode;
+  *filesize = fle.size;
+
   gftp_file_destroy (&fle, 0);
 
   return (0);
@@ -1798,7 +1815,7 @@ sshv2_open_file (gftp_request * request, const char *file, off_t startsize,
   params = request->protocol_data;
   params->offset = startsize;
 
-  len = 8;
+  len = 8; /* For mode */
   tempstr = sshv2_initialize_string_with_path (request, file, &len, &endpos);
 
   num = htonl (mode);
@@ -1836,12 +1853,11 @@ sshv2_open_file (gftp_request * request, const char *file, off_t startsize,
 
 
 static off_t
-sshv2_get_file (gftp_request * request, const char *file, int fd,
-                off_t startsize)
+sshv2_get_file (gftp_request * request, const char *file, off_t startsize)
 {
   int ret;
 
-  if ((ret = sshv2_open_file (request, file, startsize, SSH_FXP_OPEN)) < 0)
+  if ((ret = sshv2_open_file (request, file, startsize, SSH_FXF_READ)) < 0)
     return (ret);
 
   return (sshv2_get_file_size (request, file));
@@ -1849,7 +1865,7 @@ sshv2_get_file (gftp_request * request, const char *file, int fd,
 
 
 static int
-sshv2_put_file (gftp_request * request, const char *file, int fd,
+sshv2_put_file (gftp_request * request, const char *file,
                 off_t startsize, off_t totalsize)
 {
   guint32 mode;
@@ -1867,31 +1883,13 @@ sshv2_put_file (gftp_request * request, const char *file, int fd,
 }
 
 
-#ifdef G_HAVE_GINT64
-
-static gint64
-sshv2_hton64 (gint64 val)
-{
-#if G_BYTE_ORDER != G_BIG_ENDIAN
-  return (GINT64_TO_BE (val));
-#else
-  return (val);
-#endif
-}
-
-#endif
-
-
 static void
 sshv2_setup_file_offset (sshv2_params * params, char *buf)
 {
   guint32 hinum, lownum;
 #ifdef G_HAVE_GINT64
-  gint64 offset;
-
-  offset = sshv2_hton64 (params->offset);
-  lownum = offset >> 32;
-  hinum = (guint32) offset;
+  hinum = htonl(params->offset >> 32);
+  lownum = htonl((guint32) params->offset);
 #else
   hinum = 0;
   lownum = htonl (params->offset);
@@ -1920,7 +1918,7 @@ sshv2_get_next_file_chunk (gftp_request * request, char *buf, size_t size)
   if (params->transfer_buffer == NULL)
     {
       params->transfer_buffer_len = params->handle_len + 12;
-      params->transfer_buffer = g_malloc (params->transfer_buffer_len);
+      params->transfer_buffer = g_malloc0 (params->transfer_buffer_len);
       memcpy (params->transfer_buffer, params->handle, params->handle_len);
     }
 
@@ -1989,7 +1987,7 @@ sshv2_put_next_file_chunk (gftp_request * request, char *buf, size_t size)
   if (params->transfer_buffer == NULL)
     {
       params->transfer_buffer_len = params->handle_len + size + 12;
-      params->transfer_buffer = g_malloc (params->transfer_buffer_len);
+      params->transfer_buffer = g_malloc0 (params->transfer_buffer_len);
       memcpy (params->transfer_buffer, params->handle, params->handle_len);
     }
 
@@ -2085,7 +2083,7 @@ sshv2_copy_param_options (gftp_request * dest_request,
   memcpy (dparms->handle, sparms->handle, sizeof (*dparms->handle));
   if (sparms->transfer_buffer != NULL)
     {
-      dparms->transfer_buffer = g_malloc (sparms->transfer_buffer_len);
+      dparms->transfer_buffer = g_malloc0 (sparms->transfer_buffer_len);
       memcpy (dparms->transfer_buffer, sparms->transfer_buffer,
               sparms->transfer_buffer_len);
     }
@@ -2148,6 +2146,7 @@ sshv2_init (gftp_request * request)
   request->need_password = 1;
   request->use_cache = 1;
   request->always_connected = 0;
+  request->use_local_encoding = 0;
   request->protocol_data = g_malloc0 (sizeof (sshv2_params));
   request->server_type = GFTP_DIRTYPE_UNIX;
 
