@@ -1,6 +1,6 @@
 /*****************************************************************************/
 /*  gftp-gtk.h - include file for the gftp gtk+ 1.2 port                     */
-/*  Copyright (C) 1998-2002 Brian Masney <masneyb@seul.org>                  */
+/*  Copyright (C) 1998-2003 Brian Masney <masneyb@gftp.org>                  */
 /*                                                                           */
 /*  This program is free software; you can redistribute it and/or modify     */
 /*  it under the terms of the GNU General Public License as published by     */
@@ -17,16 +17,67 @@
 /*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                */
 /*****************************************************************************/
 
+/* $Id: gftp-gtk.h 920 2007-04-18 23:38:55Z masneyb $ */
+
 #ifndef __GFTP_GTK_H
 #define __GFTP_GTK_H
 
 #include "../../lib/gftp.h"
+#include "../uicommon/gftpui.h"
 #include <gtk/gtk.h>
 #include <gdk/gdkkeysyms.h>
 #include <pthread.h>
 
+#define GFTP_MENU_ITEM_ASCII	1
+#define GFTP_MENU_ITEM_BINARY	2
+#define GFTP_MENU_ITEM_WIN1	3
+#define GFTP_MENU_ITEM_WIN2	4
+
 #define IS_ONE_SELECTED(wdata)		(GTK_CLIST ((wdata)->listbox)->selection && GTK_CLIST ((wdata)->listbox)->selection->next == NULL)
 #define IS_NONE_SELECTED(wdata)		(GTK_CLIST ((wdata)->listbox)->selection == NULL)
+#define gftp_gtk_get_list_selection(wdata)  (GTK_CLIST ((wdata)->listbox)->selection)
+
+#define GFTP_IS_SAME_HOST_START_TRANS(wdata,trequest) \
+  ((wdata) != NULL && (wdata)->request != NULL && \
+  (wdata)->request->datafd > 0 && !(wdata)->request->always_connected && \
+  !(wdata)->request->stopable && \
+  compare_request (trequest, (wdata)->request, 0))
+
+#define GFTP_IS_SAME_HOST_STOP_TRANS(wdata,trequest) \
+  ((wdata) != NULL && (wdata)->request != NULL && \
+  (wdata)->request->datafd < 0 && !(wdata)->request->always_connected && \
+  (wdata)->request->cached && !(wdata)->request->stopable && \
+  trequest->datafd > 0 && !trequest->always_connected && \
+  compare_request (trequest, (wdata)->request, 0))
+
+#if GTK_MAJOR_VERSION == 1
+  #define gtk_widget_set_size_request(widget, width, height)	\
+				gtk_widget_set_usize (widget, width, height)
+#endif
+
+/* These 2 defines are for creating menu items with stock icons in GTK+ 2.0. 
+   If we are using version 1.2, it will disable the stock items since it's not
+   supported */
+
+#if GTK_MAJOR_VERSION < 2
+#define MS_(a) NULL
+#define MN_(a) a
+#else
+#define MS_(a) "<StockItem>",a
+#define MN_(a) a,NULL
+#endif
+
+/* These are used for the MakeEditDialog function. I have these types to make
+   it easier for creating dialogs with GTK+ 1.2 and GTK+ 2.0 */
+
+typedef enum gftp_dialog_button_tag
+{
+  gftp_dialog_button_create,
+  gftp_dialog_button_change,
+  gftp_dialog_button_connect,
+  gftp_dialog_button_rename,
+  gftp_dialog_button_ok
+} gftp_dialog_button;
 
 typedef struct gftp_window_data_tag
 {
@@ -34,9 +85,8 @@ typedef struct gftp_window_data_tag
    				   a directory */
             *hoststxt, 		/* Show which directory we're in */
             *listbox; 		/* Our listbox showing the files */
-  unsigned int sortcol, 	/* Which column we are sorting by */
-               sortasds : 1, 	/* Sorted ascending or descending */
-               sorted : 1,	/* Is the output sorted? */
+  unsigned int sorted : 1,	/* Is the output sorted? */
+               show_selected : 1, /* Show only selected files */
                *histlen;	/* Pointer to length of history */
   char *filespec;		/* Filespec for the listbox */
   gftp_request * request;	/* The host that we are connected to */
@@ -45,10 +95,17 @@ typedef struct gftp_window_data_tag
   GtkItemFactory *ifactory; 	/* This is for the menus that will
                                    come up when you right click */
   pthread_t tid;		/* Thread for the stop button */
-  unsigned long gotbytes;
+  char *prefix_col_str;
 } gftp_window_data;
 
 
+typedef struct _gftpui_gtk_thread_data
+{
+  void * (*func) (void *);
+  gftpui_callback_data * cdata;
+} gftpui_gtk_thread_data;
+
+
 typedef struct gftp_graphic_tag
 {
   char * filename;
@@ -59,11 +116,15 @@ typedef struct gftp_graphic_tag
 
 typedef struct gftp_dialog_data_tag
 {
-  GtkWidget * dialog,		/* Pointer to the dialog */
-            * checkbox,		/* Pointer to the checkbox widget */
-            * edit; 		/* Pointer to the edit widget */
-  GList * all_buttons;		/* All the buttons of this dialog */
-  gpointer data;		/* Data pointer we'll pass to the function */
+  GtkWidget * dialog,
+            * checkbox,
+            * edit;
+
+  void (*yesfunc) ();
+  gpointer yespointer;
+
+  void (*nofunc) ();
+  gpointer nopointer;
 } gftp_dialog_data;
 
 
@@ -82,25 +143,58 @@ typedef struct gftp_viewedit_data_tag
 				   editing it */
    gftp_window_data * fromwdata, /* The window we are viewing this file in */
                     * towdata;
+   gftp_request * torequest;
 } gftp_viewedit_data;
 
 
-typedef struct gftp_curtrans_data_tag
+typedef struct gftp_save_dir_struct_tag
 {
-  gftp_transfer * transfer;
-  GList * curfle;
-} gftp_curtrans_data;
+  GtkWidget * filew;
+  gftp_window_data * wdata;
+} gftp_save_dir_struct;
+
+
+typedef struct gftp_textcomboedt_widget_data_tag
+{
+  GtkWidget * combo,
+            * text;
+  gftp_config_vars * cv;
+  char * custom_edit_value;
+} gftp_textcomboedt_widget_data;
+
+
+typedef struct gftp_options_dialog_data_tag
+{
+  GtkWidget * dialog,
+            * notebook,
+            * box,
+            * table;
+  unsigned int tbl_col_num,
+               tbl_row_num;
+  gftp_option_type_enum last_option;
+  gftp_bookmarks_var * bm;
+} gftp_options_dialog_data;
 
 
 extern gftp_window_data window1, window2, * other_wdata, * current_wdata;
 extern GtkWidget * stop_btn, * hostedit, * useredit, * passedit,
-                 * portedit, * logwdw, * dlwdw, * protocol_menu, * optionmenu;
+                 * portedit, * logwdw, * dlwdw, * optionmenu,
+                 * gftpui_command_widget, * download_left_arrow,
+                 * upload_right_arrow, * openurl_btn;
+extern GtkTooltips * openurl_tooltip;
 extern GtkAdjustment * logwdw_vadj;
-extern int local_start, remote_start, trans_start, log_start, tools_start;
+#if GTK_MAJOR_VERSION > 1
+extern GtkTextMark * logwdw_textmark;
+#endif
+extern int local_start, remote_start, trans_start;
 extern GHashTable * graphic_hash_table;
 extern GtkItemFactoryEntry * menus;
 extern GtkItemFactory * factory;
-extern pthread_mutex_t transfer_mutex, log_mutex;
+extern pthread_mutex_t log_mutex;
+extern gftp_graphic * gftp_icon;
+extern pthread_t main_thread_id;
+extern GList * viewedit_processes;
+
 
 /* bookmarks.c */
 void run_bookmark 				( gpointer data );
@@ -144,6 +238,8 @@ void listbox_get_drag_data 			( GtkWidget * widget,
 						  gpointer data );
 
 /* gftp-gtk.c */
+void gftp_gtk_init_request 			( gftp_window_data * wdata );
+
 void toolbar_hostedit 				( GtkWidget * widget, 
 						  gpointer data );
 
@@ -154,32 +250,47 @@ void sortrows 					( GtkCList * clist,
 void stop_button				( GtkWidget * widget,
 						  gpointer data );
 
+void gftpui_show_or_hide_command 		( void );
+
+/* gtkui.c */
+void gftpui_run_command 			( GtkWidget * widget,
+						  gpointer data );
+
+void gftpui_run_function_callback 		( gftp_window_data * wdata,
+						  gftp_dialog_data * ddata );
+
+void gftpui_run_function_cancel_callback 	( gftp_window_data * wdata,
+						  gftp_dialog_data * ddata );
+
+void gftpui_mkdir_dialog 			( gpointer data );
+
+void gftpui_rename_dialog			( gpointer data );
+
+void gftpui_site_dialog 			( gpointer data );
+
+int gftpui_run_chdir 				( gpointer uidata,
+						  char *directory );
+
+void gftpui_chdir_dialog 			( gpointer data );
+
+char * gftpui_gtk_get_utf8_file_pos 		( gftp_file * fle );
+
 /* menu_items.c */
-void change_setting 				( gftp_window_data *wdata,
-						  int menuitem,
-						  GtkWidget * checkmenu );
-
-void tb_openurl_dialog 				( gpointer data );
-
-void openurl_dialog 				( gpointer data );
-
-void disconnect 				( gpointer data );
-
 void change_filespec 				( gpointer data );
 
+void save_directory_listing 			( gpointer data );
+
+void show_selected				( gpointer data );
+
 void selectall 					( gpointer data );
 
 void selectallfiles 				( gpointer data );
 
 void deselectall 				( gpointer data );
 
-void site_dialog 				( gpointer data );
-
 int chdir_edit					( GtkWidget * widget,
 						  gpointer data );
 
-int chdir_dialog 				( gpointer data );
-
 void clearlog 					( gpointer data );
 
 void viewlog 					( gpointer data );
@@ -188,9 +299,9 @@ void savelog 					( gpointer data );
 
 void clear_cache				( gpointer data );
 
-void options_dialog (gpointer data);
-void compare_windows (gpointer data);
-void about_dialog (gpointer data);
+void compare_windows 				( gpointer data );
+
+void about_dialog 				( gpointer data );
 
 /* misc-gtk.c */
 void fix_display				( void );
@@ -198,11 +309,9 @@ void fix_display				( void );
 void remove_files_window			( gftp_window_data * wdata );
 
 void ftp_log					( gftp_logging_level level,
-						  void *ptr,
+						  gftp_request * request,
 						  const char *string,
-						  ... );
-
-void refresh 					( gftp_window_data * wdata );
+						  ... ) GFTP_LOG_FUNCTION_ATTRIBUTES;
 
 void update_window_info				( void );
 
@@ -223,20 +332,21 @@ void gftp_get_pixmap 				( GtkWidget * widget,
 
 int check_status				( char *name,
 						  gftp_window_data * wdata,
-						  int check_other_stop,
-						  int only_one,
-						  int at_least_one,
-						  int func );
+						  unsigned int check_other_stop,
+						  unsigned int only_one,
+						  unsigned int at_least_one,
+						  unsigned int func );
+
+GtkItemFactory *item_factory_new                ( GtkType	       container_type,
+						  const char	      *path,
+						  GtkAccelGroup       *accel_group,
+						  const char          *strip_prefix );
 
 void create_item_factory 			( GtkItemFactory * ifactory, 
-						  guint n_entries, 
+						  gint n_entries, 
 						  GtkItemFactoryEntry * entries,
 						  gpointer callback_data );
 
-GList * get_next_selection 			( GList * selection, 
-						  GList ** list, 
-						  int *curnum );
-
 void add_history 				( GtkWidget * widget, 
 						  GList ** history, 
 						  unsigned int *histlen, 
@@ -247,53 +357,51 @@ int check_reconnect 				( gftp_window_data * wdata );
 void add_file_listbox 				( gftp_window_data * wdata, 
 						  gftp_file * fle );
 
+void destroy_dialog 				( gftp_dialog_data * ddata );
+
 void MakeEditDialog 				( char *diagtxt, 
 						  char *infotxt, 
 						  char *deftext, 
-						  int edit_shown, 
-						  int erase, 
-                                                  char *checktext,
-						  char *oktxt, 
+						  int passwd_item,
+						  char *checktext, 
+						  gftp_dialog_button okbutton, 
 						  void (*okfunc) (), 
 						  void *okptr, 
-						  char *canceltxt, 
 						  void (*cancelfunc) (), 
 						  void *cancelptr );
 
 void MakeYesNoDialog 				( char *diagtxt, 
 						  char *infotxt, 
-						  int erase, 
-						  int num, 
-						  ... );
+						  void (*yesfunc) (), 
+						  gpointer yespointer,
+						  void (*nofunc) (), 
+						  gpointer nopointer );
 
 void update_directory_download_progress 	( gftp_transfer * transfer );
 
-void *generic_thread 				( void * (*func) 
-                                                         (void *), 
-						  gftp_window_data * wdata );
-
 int progress_timeout 				( gpointer data );
 
 void display_cached_logs			( void );
 
-/* mkdir_dialog.c */
-void mkdir_dialog 				( gpointer data );
+char * get_xpm_path 				( char *filename, 
+						  int quit_on_err );
 
-/* rename_dialog.c */
-void rename_dialog				( gpointer data );
+/* options_dialog.c */
+void options_dialog 				( gpointer data );
+
+void gftp_gtk_setup_bookmark_options 		( GtkWidget * notebook,
+						  gftp_bookmarks_var * bm );
+
+void gftp_gtk_save_bookmark_options 		( void );
 
 /* transfer.c */
-int ftp_list_files				( gftp_window_data * wdata,
-						  int usecache );
+int ftp_list_files				( gftp_window_data * wdata );
 
 int ftp_connect					( gftp_window_data * wdata,
-						  gftp_request * request,
-						  int getdir );
+						  gftp_request * request );
 
 gint update_downloads 				( gpointer data );
 
-void * gftp_gtk_transfer_files 			( void *data );
-
 void get_files 					( gpointer data );
 
 void put_files 					( gpointer data );
@@ -301,14 +409,9 @@ void put_files 					( gpointer data );
 void transfer_window_files 			( gftp_window_data * fromwdata,
 						  gftp_window_data * towdata );
 
-void *do_getdir_thread 				( void * data );
+int gftp_gtk_get_subdirs 			( gftp_transfer * transfer );
 
-void add_file_transfer 				( gftp_request * fromreq, 
-						  gftp_request * toreq, 
-						  gftp_window_data * fromwdata,
-						  gftp_window_data * towdata, 
-						  GList * files,
-						  int copy_req );
+void *do_getdir_thread 				( void * data );
 
 void start_transfer				( gpointer data );
 
@@ -322,8 +425,6 @@ void move_transfer_up				( gpointer data );
 
 void move_transfer_down				( gpointer data );
 
-void gftp_gtk_ask_transfer 			( gftp_transfer * tdata );
-
 /* view_dialog.c */
 void edit_dialog 				( gpointer data );
 
@@ -331,10 +432,10 @@ void view_dialog 				( gpointer data );
 
 void view_file 					( char *filename, 
 						  int fd, 
-						  int viewedit, 
-						  int del_file, 
-						  int start_pos, 
-						  int dontupload,
+						  unsigned int viewedit, 
+						  unsigned int del_file, 
+						  unsigned int start_pos, 
+						  unsigned int dontupload,
 						  char *remote_filename, 
 						  gftp_window_data * wdata );
 
