@@ -1,6 +1,6 @@
 /*****************************************************************************/
 /*  menu-items.c - menu callbacks                                            */
-/*  Copyright (C) 1998-2002 Brian Masney <masneyb@seul.org>                  */
+/*  Copyright (C) 1998-2007 Brian Masney <masneyb@gftp.org>                  */
 /*                                                                           */
 /*  This program is free software; you can redistribute it and/or modify     */
 /*  it under the terms of the GNU General Public License as published by     */
@@ -17,182 +17,18 @@
 /*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.                */
 /*****************************************************************************/
 
-#include <gftp-gtk.h>
-
-static void do_openurl 				( GtkWidget * widget, 
-						  gftp_dialog_data * data );
-static void dochange_filespec 			( GtkWidget * widget, 
-						  gftp_dialog_data * data );
-static void dosite 				( GtkWidget * widget, 
-						  gftp_dialog_data * data );
-static int do_change_dir			( gftp_window_data * wdata,
-						  char * directory );
-static void *do_change_dir_thread 		( void * data );
-static void dosavelog 				( GtkWidget * widget, 
-						  GtkFileSelection * fs );
-
-static RETSIGTYPE sig_chdirquit                 ( int signo );
-static sigjmp_buf chdirenvir;
-
-void
-change_setting (gftp_window_data * wdata, int menuitem, GtkWidget * checkmenu)
-{
-  switch (menuitem)
-    {
-    case 1:
-      if (wdata->request != NULL)
-        gftp_set_data_type (wdata->request, GFTP_TYPE_ASCII);
-      break;
-    case 2:
-      if (wdata->request != NULL)
-        gftp_set_data_type (wdata->request, GFTP_TYPE_BINARY);
-      break;
-    case 3:
-      current_wdata = &window1;
-      other_wdata = &window2;
-      if (wdata->request)
-        update_window_info ();
-      break;
-    case 4:
-      current_wdata = &window2;
-      other_wdata = &window1;
-      if (wdata->request)
-        update_window_info ();
-      break;
-    }
-}
-
-
-void
-tb_openurl_dialog (gpointer data)
-{
-  const char *edttxt;
-
-  if (current_wdata->request->stopable)
-    {
-      ftp_log (gftp_logging_misc, NULL,
-               _("%s: Please hit the stop button first to do anything else\n"),
-               _("OpenURL"));
-      return;
-    }
-
-  edttxt = gtk_entry_get_text (GTK_ENTRY (GTK_COMBO (hostedit)->entry));
-  if (*edttxt == '\0')
-    {
-      ftp_log (gftp_logging_misc, NULL,
-               _("OpenURL: Operation canceled...you must enter a string\n"));
-      return;
-    }
-
-  if (GFTP_IS_CONNECTED (current_wdata->request))
-    disconnect (current_wdata);
-  else if (edttxt != NULL && *edttxt != '\0')
-    toolbar_hostedit (NULL, NULL);
-  else
-    openurl_dialog (current_wdata);
-}
-
-
-void
-openurl_dialog (gpointer data)
-{
-  MakeEditDialog (_("Connect via URL"), _("Enter ftp url to connect to"),
-                  NULL, 1, 1, NULL, _("Connect"), do_openurl, data,
-                  _("  Cancel  "), NULL, NULL);
-}
-
+#include "gftp-gtk.h"
+static const char cvsid[] = "$Id: menu-items.c 949 2008-01-24 23:26:45Z masneyb $";
 
 static void
-do_openurl (GtkWidget * widget, gftp_dialog_data * data)
-{
-  gftp_window_data * wdata;
-  const char *tempstr;
-
-  wdata = data->data;
-  if (current_wdata->request->stopable)
-    {
-      ftp_log (gftp_logging_misc, NULL,
-               _("%s: Please hit the stop button first to do anything else\n"),
-               _("OpenURL"));
-      return;
-    }
-
-  tempstr = gtk_entry_get_text (GTK_ENTRY (data->edit));
-  if (*tempstr == '\0')
-    {
-      ftp_log (gftp_logging_misc, NULL,
-               _("OpenURL: Operation canceled...you must enter a string\n"));
-      return;
-    }
-
-  if (GFTP_IS_CONNECTED (wdata->request))
-    disconnect (wdata);
-
-  if (gftp_parse_url (wdata->request, tempstr) == 0)
-    {
-      gtk_widget_destroy (data->dialog);
-      ftp_connect (wdata, wdata->request, 1);
-    }
-  else
-    {
-      gtk_widget_destroy (data->dialog);
-      ftp_log (gftp_logging_misc, NULL, _("Could not parse URL %s\n"), tempstr);    
-    }
-  data->dialog = NULL; 
-}
-
-
-void 
-disconnect (gpointer data)
-{
-  gftp_window_data * wdata;
-
-  wdata = data;
-  gftp_disconnect (wdata->request);
-  remove_files_window (wdata);
-  update_window (wdata);
-}
-
-
-void 
-change_filespec (gpointer data)
-{
-  gftp_window_data * wdata;
-
-  wdata = data;
-  if (!check_status (_("Change Filespec"), wdata, 0, 0, 0, 1))
-    return;
-
-  MakeEditDialog (_("Change Filespec"), _("Enter the new file specification"),
-                  wdata->filespec, 1, 1, NULL, _("Change"), dochange_filespec,
-                  wdata, _("  Cancel  "), NULL, NULL);
-}
-
-
-static void
-dochange_filespec (GtkWidget * widget, gftp_dialog_data * data)
+update_window_listbox (gftp_window_data * wdata)
 {
   GList * templist, * filelist;
   gftp_file * tempfle;
-  gftp_window_data * wdata;
-  const char *edttext;
   int num;
 
-  wdata = data->data;
-  edttext = gtk_entry_get_text (GTK_ENTRY (data->edit));
-  if (*edttext == '\0')
-    {
-      ftp_log (gftp_logging_misc, NULL,
-               _("Change Filespec: Operation canceled...you must enter a string\n"));
-      return;
-    }
-  if (wdata->filespec)
-    g_free (wdata->filespec);
-  wdata->filespec = g_malloc (strlen (edttext) + 1);
-  strcpy (wdata->filespec, edttext);
-
   filelist = wdata->files;
-  templist = GTK_CLIST (wdata->listbox)->selection;
+  templist = gftp_gtk_get_list_selection (wdata);
   num = 0;
   while (templist != NULL)
     {
@@ -215,12 +51,134 @@ dochange_filespec (GtkWidget * widget, gftp_dialog_data * data)
 }
 
 
+static void
+dochange_filespec (gftp_window_data * wdata, gftp_dialog_data * ddata)
+{
+  const char *edttext;
+
+  wdata->show_selected = 0;
+
+  edttext = gtk_entry_get_text (GTK_ENTRY (ddata->edit));
+  if (*edttext == '\0')
+    {
+      ftp_log (gftp_logging_error, NULL,
+               _("Change Filespec: Operation canceled...you must enter a string\n"));
+      return;
+    }
+
+  if (wdata->filespec)
+    g_free (wdata->filespec);
+
+  wdata->filespec = g_strdup (edttext);
+  update_window_listbox (wdata);
+}
+
+
+void 
+change_filespec (gpointer data)
+{
+  gftp_window_data * wdata;
+
+  wdata = data;
+  if (!check_status (_("Change Filespec"), wdata, 0, 0, 0, 1))
+    return;
+
+  MakeEditDialog (_("Change Filespec"), _("Enter the new file specification"),
+                  wdata->filespec, 1, NULL, gftp_dialog_button_change, 
+                  dochange_filespec, wdata, NULL, NULL);
+}
+
+
+static void
+destroy_save_directory_listing (GtkWidget * widget, gftp_save_dir_struct * str)
+{
+  gtk_widget_destroy (str->filew);
+  g_free (str);
+}
+
+
+static void
+dosave_directory_listing (GtkWidget * widget, gftp_save_dir_struct * str)
+{
+  const char *filename;
+  gftp_file * tempfle;
+  GList * templist;
+  char *tempstr;
+  FILE * fd;
+ 
+
+  filename = gtk_file_selection_get_filename (GTK_FILE_SELECTION (str->filew));
+  if ((fd = fopen (filename, "w")) == NULL)
+    {
+      ftp_log (gftp_logging_error, NULL, 
+               _("Error: Cannot open %s for writing: %s\n"), filename, 
+               g_strerror (errno));
+      return;
+    }
+
+  for (templist = str->wdata->files; 
+       templist != NULL;
+       templist = templist->next)
+    {
+      tempfle = templist->data;
+
+      if (!tempfle->shown)
+        continue;
+
+      tempstr = gftp_gen_ls_string (NULL, tempfle, NULL, NULL);
+      fprintf (fd, "%s\n", tempstr);
+      g_free (tempstr);
+    }
+
+  fclose (fd);
+}
+
+
+void 
+save_directory_listing (gpointer data)
+{
+  gftp_save_dir_struct * str;
+  GtkWidget *filew;
+
+  filew = gtk_file_selection_new (_("Save Directory Listing"));
+
+  str = g_malloc0 (sizeof (*str));
+  str->filew = filew;
+  str->wdata = data;
+
+  gtk_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION (filew)->ok_button),
+                      "clicked", GTK_SIGNAL_FUNC (dosave_directory_listing), 
+                      str);
+  gtk_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION (filew)->ok_button),
+                      "clicked", 
+                      GTK_SIGNAL_FUNC (destroy_save_directory_listing), str);
+  gtk_signal_connect (GTK_OBJECT (GTK_FILE_SELECTION (filew)->cancel_button), 
+                      "clicked", 
+                      GTK_SIGNAL_FUNC (destroy_save_directory_listing), str);
+
+  gtk_window_set_wmclass (GTK_WINDOW(filew), "Save Directory Listing", "gFTP");
+  gtk_widget_show (filew);
+}
+
+
+void
+show_selected (gpointer data)
+{
+  gftp_window_data * wdata;
+
+  wdata = data;
+  wdata->show_selected = 1;
+  update_window_listbox (wdata);
+}
+
+
 void 
 selectall (gpointer data)
 {
   gftp_window_data * wdata;
 
   wdata = data;
+  wdata->show_selected = 0;
   gtk_clist_select_all (GTK_CLIST (wdata->listbox));
 }
 
@@ -234,15 +192,16 @@ selectallfiles (gpointer data)
   int i;
 
   wdata = data;
+  wdata->show_selected = 0;
   gtk_clist_freeze (GTK_CLIST (wdata->listbox));
   i = 0;
-  templist = g_list_first (wdata->files);
+  templist = wdata->files;
   while (templist != NULL)
     {
       tempfle = (gftp_file *) templist->data;
       if (tempfle->shown)
 	{
-	  if (tempfle->isdir)
+	  if (S_ISDIR (tempfle->st_mode))
 	    gtk_clist_unselect_row (GTK_CLIST (wdata->listbox), i, 0);
 	  else
 	    gtk_clist_select_row (GTK_CLIST (wdata->listbox), i, 0);
@@ -260,48 +219,11 @@ deselectall (gpointer data)
   gftp_window_data * wdata;
 
   wdata = data;
+  wdata->show_selected = 0;
   gtk_clist_unselect_all (GTK_CLIST (wdata->listbox));
 }
 
 
-void
-site_dialog (gpointer data)
-{
-  gftp_window_data * wdata;
-
-  wdata = data;
-  if (!check_status (_("Site"), wdata, 0, 0, 0, wdata->request->site != NULL))
-    return;
-
-  MakeEditDialog (_("Site"), _("Enter site-specific command"), NULL, 1, 1,
-                  NULL, _("OK"), dosite, wdata, _("  Cancel  "), NULL, NULL);
-}
-
-
-static void
-dosite (GtkWidget * widget, gftp_dialog_data * data)
-{
-  gftp_window_data * wdata;
-  const char *edttext;
-
-  wdata = data->data;
-  edttext = gtk_entry_get_text (GTK_ENTRY (data->edit));
-  if (*edttext == '\0')
-    {
-      ftp_log (gftp_logging_misc, NULL,
-               _("SITE: Operation canceled...you must enter a string\n"));
-      return;
-    }
-
-  if (check_reconnect (wdata) < 0)
-    return;
-  gftp_site_cmd (wdata->request, edttext);
-
-  if (!GFTP_IS_CONNECTED (wdata->request))
-    disconnect (wdata);
-}
-
-
 int
 chdir_edit (GtkWidget * widget, gpointer data)
 {
@@ -310,175 +232,38 @@ chdir_edit (GtkWidget * widget, gpointer data)
   char *tempstr;
 
   wdata = data;
+  if (!check_status (_("Chdir"), wdata, gftpui_common_use_threads (wdata->request), 0, 0,
+                     wdata->request->chdir != NULL))
+    return (0);
+
+  if (check_reconnect (wdata) < 0)
+    return (0);
+
   edttxt = gtk_entry_get_text (GTK_ENTRY (GTK_COMBO (wdata->combo)->entry));
+
   if (!GFTP_IS_CONNECTED (wdata->request) && *edttxt != '\0')
     {
       toolbar_hostedit (NULL, NULL);
       return (0);
     }
 
-  if (!check_status (_("Chdir"), wdata, wdata->request->use_threads, 0, 0, 
-                     wdata->request->chdir != NULL))
+  if ((tempstr = gftp_expand_path (wdata->request, edttxt)) == NULL)
     return (FALSE);
 
-  if ((tempstr = expand_path (edttxt)) == NULL)
-    return (FALSE);
-
-  if (check_reconnect (wdata) < 0)
-    return (FALSE);
-
-  if (do_change_dir (wdata, tempstr) == 0)
-    {
-      gtk_clist_freeze (GTK_CLIST (wdata->listbox));
-      remove_files_window (wdata);
-      ftp_list_files (wdata, 1);
-      gtk_clist_thaw (GTK_CLIST (wdata->listbox));
-      add_history (wdata->combo, wdata->history, wdata->histlen, tempstr);
-    }
+  if (gftpui_run_chdir (wdata, tempstr))
+    add_history (wdata->combo, wdata->history, wdata->histlen, edttxt);
 
   g_free (tempstr);
-  return (FALSE);
-}
-
-
-int
-chdir_dialog (gpointer data)
-{
-  GList * templist, * filelist;
-  gftp_window_data * wdata;
-  char *newdir, *tempstr;
-  gftp_file *tempfle;
-  int num, ret;
-
-  wdata = data;
-  if (!check_status (_("Chdir"), wdata, wdata->request->use_threads, 1, 0, 
-                     wdata->request->chdir != NULL))
-    return (0);
-
-  filelist = g_list_first (wdata->files);
-  templist = GTK_CLIST (wdata->listbox)->selection;
-  num = 0;
-  templist = get_next_selection (templist, &filelist, &num);
-  tempfle = filelist->data;
-
-  newdir = g_strconcat (GFTP_GET_DIRECTORY (wdata->request), "/",
-		        tempfle->file, NULL);
-  remove_double_slashes (newdir);
-  if ((tempstr = expand_path (newdir)) == NULL)
-    return (0);
-  g_free (newdir);
-
-  if (check_reconnect (wdata) < 0)
-    return (0);
-
-  ret = 0; 
-  if (do_change_dir (wdata, tempstr) == 0)
-    {
-      gtk_clist_freeze (GTK_CLIST (wdata->listbox));
-      remove_files_window (wdata);
-      ftp_list_files (wdata, 1);
-      gtk_clist_thaw (GTK_CLIST (wdata->listbox));
-      ret = 1;
-    }
-  g_free (tempstr);
-  return (ret);
-}
-
-
-static int
-do_change_dir (gftp_window_data * wdata, char *directory)
-{
-  char *olddir;
-  int ret;
-
-  if (directory != wdata->request->directory)
-    {
-      olddir = wdata->request->directory;
-      wdata->request->directory = g_malloc (strlen (directory) + 1);
-      strcpy (wdata->request->directory, directory);
-    }
-  else
-    olddir = NULL;
-
-  ret = (int) generic_thread (do_change_dir_thread, wdata);
-
-  if (!GFTP_IS_CONNECTED (wdata->request))
-    {
-      disconnect (wdata);
-      if (olddir != NULL)
-        g_free (olddir);
-      return (-2);
-    }
-
-  if (ret != 0)
-    {
-      g_free (wdata->request->directory);
-      wdata->request->directory = olddir;
-    }
-
-  return (ret);
-}
-
-
-static void *
-do_change_dir_thread (void * data)
-{
-  gftp_window_data * wdata;
-  int success, sj;
-
-  wdata = data;
-  wdata->request->user_data = (void *) 0x01;
-
-  if (wdata->request->use_threads)
-    {
-      sj = sigsetjmp (chdirenvir, 1);
-      signal (SIGINT, sig_chdirquit);
-      signal (SIGALRM, sig_chdirquit);
-    }
-  else
-    sj = 0;
-
-  success = 0;
-  if (sj == 0) 
-    {
-      if (wdata->request->connect_timeout > 0)
-        alarm (wdata->request->connect_timeout);
-      success = gftp_set_directory (wdata->request, wdata->request->directory);
-      alarm (0);
-    }
-  else
-    {
-      gftp_disconnect (wdata->request);
-      wdata->request->logging_function (gftp_logging_error,
-                                        wdata->request->user_data,
-                                        _("Operation canceled\n"));
-    }
-
-  if (wdata->request->use_threads)
-    {
-      signal (SIGINT, SIG_DFL);
-      signal (SIGALRM, SIG_IGN);
-    }
-
-  wdata->request->user_data = NULL;
-  wdata->request->stopable = 0;
-  return ((void *) success);
-}
-
-
-static RETSIGTYPE
-sig_chdirquit (int signo)
-{
-  signal (signo, sig_chdirquit);
-  siglongjmp (chdirenvir, signo == SIGINT ? 1 : 2);
+  return (0);
 }
 
 
 void 
 clearlog (gpointer data)
 {
-  guint len;
-#if GTK_MAJOR_VERSION == 1 && GTK_MINOR_VERSION == 2
+  gint len;
+
+#if GTK_MAJOR_VERSION == 1
   len = gtk_text_get_length (GTK_TEXT (logwdw));
   gtk_text_set_point (GTK_TEXT (logwdw), len);
   gtk_text_backward_delete (GTK_TEXT (logwdw), len);
@@ -489,7 +274,7 @@ clearlog (gpointer data)
   textbuf = gtk_text_view_get_buffer (GTK_TEXT_VIEW (logwdw));
   len = gtk_text_buffer_get_char_count (textbuf);
   gtk_text_buffer_get_iter_at_offset (textbuf, &iter, 0);
-  gtk_text_buffer_get_iter_at_offset (textbuf, &iter2, len - 1);
+  gtk_text_buffer_get_iter_at_offset (textbuf, &iter2, len);
   gtk_text_buffer_delete (textbuf, &iter, &iter2);
 #endif
 }
@@ -499,51 +284,124 @@ void
 viewlog (gpointer data)
 {
   char *tempstr, *txt, *pos;
-  guint textlen;
+  gint textlen;
   ssize_t len;
   int fd;
-#if !(GTK_MAJOR_VERSION == 1 && GTK_MINOR_VERSION == 2)
+#if GTK_MAJOR_VERSION > 1
   GtkTextBuffer * textbuf;
   GtkTextIter iter, iter2;
 #endif
 
-  tempstr = g_strconcat (tmp_directory, "/gftp-view.XXXXXXXXXX", NULL);
+  tempstr = g_strconcat (g_get_tmp_dir (), "/gftp-view.XXXXXXXXXX", NULL);
   if ((fd = mkstemp (tempstr)) < 0)
     {
-      ftp_log (gftp_logging_misc, NULL, _("Error: Cannot open %s: %s\n"),
-	       tempstr, g_strerror (errno));
+      ftp_log (gftp_logging_error, NULL, 
+               _("Error: Cannot open %s for writing: %s\n"), tempstr, 
+               g_strerror (errno));
       g_free (tempstr); 
       return;
     }
   chmod (tempstr, S_IRUSR | S_IWUSR);
-  unlink (tempstr);
   
-#if GTK_MAJOR_VERSION == 1 && GTK_MINOR_VERSION == 2
+#if GTK_MAJOR_VERSION == 1
   textlen = gtk_text_get_length (GTK_TEXT (logwdw));
   txt = gtk_editable_get_chars (GTK_EDITABLE (logwdw), 0, -1);
 #else
   textbuf = gtk_text_view_get_buffer (GTK_TEXT_VIEW (logwdw));
   textlen = gtk_text_buffer_get_char_count (textbuf);
   gtk_text_buffer_get_iter_at_offset (textbuf, &iter, 0);
-  gtk_text_buffer_get_iter_at_offset (textbuf, &iter2, textlen - 1);
+  gtk_text_buffer_get_iter_at_offset (textbuf, &iter2, textlen);
   txt = gtk_text_buffer_get_text (textbuf, &iter, &iter2, 0);
+
+  /* gtk_text_buffer_get_char_count() returns the number of characters,
+     not bytes. So get the number of bytes that need to be written out */
+  textlen = strlen (txt);
 #endif
   pos = txt;
-  do 
+
+  while (textlen > 0)
     {
       if ((len = write (fd, pos, textlen)) == -1)
         { 
-          ftp_log (gftp_logging_misc, NULL, _("Error: Error writing to %s: %s\n"), tempstr, g_strerror (errno));
+          ftp_log (gftp_logging_error, NULL, 
+                   _("Error: Error writing to %s: %s\n"), 
+                   tempstr, g_strerror (errno));
           break;
         }
       textlen -= len;
       pos += len;
+    }
+
+  fsync (fd);
+  lseek (fd, 0, SEEK_SET);
+  view_file (tempstr, fd, 1, 1, 0, 1, NULL, NULL);
+  close (fd);
+
+  g_free (tempstr);
+  g_free (txt);
+}
+
+
+static void
+dosavelog (GtkWidget * widget, GtkFileSelection * fs)
+{
+  const char *filename;
+  char *txt, *pos;
+  gint textlen;
+  ssize_t len;
+  FILE *fd;
+  int ok;
+#if GTK_MAJOR_VERSION > 1
+  GtkTextBuffer * textbuf;
+  GtkTextIter iter, iter2;
+#endif
+
+  filename = gtk_file_selection_get_filename (GTK_FILE_SELECTION (fs));
+  if ((fd = fopen (filename, "w")) == NULL)
+    {
+      ftp_log (gftp_logging_error, NULL, 
+               _("Error: Cannot open %s for writing: %s\n"), filename, 
+               g_strerror (errno));
+      return;
+    }
+
+#if GTK_MAJOR_VERSION == 1
+  textlen = gtk_text_get_length (GTK_TEXT (logwdw));
+  txt = gtk_editable_get_chars (GTK_EDITABLE (logwdw), 0, -1);
+#else
+  textbuf = gtk_text_view_get_buffer (GTK_TEXT_VIEW (logwdw));
+  textlen = gtk_text_buffer_get_char_count (textbuf);
+  gtk_text_buffer_get_iter_at_offset (textbuf, &iter, 0);
+  gtk_text_buffer_get_iter_at_offset (textbuf, &iter2, textlen);
+  txt = gtk_text_buffer_get_text (textbuf, &iter, &iter2, 0);
+
+  /* gtk_text_buffer_get_char_count() returns the number of characters,
+     not bytes. So get the number of bytes that need to be written out */
+  textlen = strlen (txt);
+#endif
+
+  ok = 1;
+  pos = txt;
+  do
+    {
+      if ((len = write (fileno (fd), pos, textlen)) == -1)
+        {
+          ok = 0;
+          ftp_log (gftp_logging_error, NULL, 
+                   _("Error: Error writing to %s: %s\n"), 
+                   filename, g_strerror (errno));
+          break;
+        }
+
+      textlen -= len;
+      pos += len;
     } while (textlen > 0);
 
-  lseek (fd, 0, SEEK_SET);
-  view_file (tempstr, fd, 1, 0, 0, 1, NULL, NULL);
-  close (fd);
-  g_free (tempstr);
+  if (ok)
+    ftp_log (gftp_logging_misc, NULL,
+             _("Successfully wrote the log file to %s\n"), filename);
+
+  fclose (fd);
   g_free (txt);
 }
 
@@ -568,63 +426,6 @@ savelog (gpointer data)
 }
 
 
-static void
-dosavelog (GtkWidget * widget, GtkFileSelection * fs)
-{
-  const char *filename;
-  char *txt, *pos;
-  guint textlen;
-  ssize_t len;
-  FILE *fd;
-  int ok;
-#if !(GTK_MAJOR_VERSION == 1 && GTK_MINOR_VERSION == 2)
-  GtkTextBuffer * textbuf;
-  GtkTextIter iter, iter2;
-#endif
-
-  filename = gtk_file_selection_get_filename (GTK_FILE_SELECTION (fs));
-  if ((fd = fopen (filename, "w")) == NULL)
-    {
-      ftp_log (gftp_logging_misc, NULL, _("Error: Cannot open %s: %s\n"),
-               filename, g_strerror (errno));
-      return;
-    }
-
-#if GTK_MAJOR_VERSION == 1 && GTK_MINOR_VERSION == 2
-  textlen = gtk_text_get_length (GTK_TEXT (logwdw));
-  txt = gtk_editable_get_chars (GTK_EDITABLE (logwdw), 0, -1);
-#else
-  textbuf = gtk_text_view_get_buffer (GTK_TEXT_VIEW (logwdw));
-  textlen = gtk_text_buffer_get_char_count (textbuf);
-  gtk_text_buffer_get_iter_at_offset (textbuf, &iter, 0);
-  gtk_text_buffer_get_iter_at_offset (textbuf, &iter2, textlen - 1);
-  txt = gtk_text_buffer_get_text (textbuf, &iter, &iter2, 0);
-#endif
-
-  ok = 1;
-  pos = txt;
-  do
-    {
-      if ((len = write (fileno (fd), pos, textlen)) == -1)
-        {
-          ok = 0;
-          ftp_log (gftp_logging_misc, NULL, _("Error: Error writing to %s: %s\n"), filename, g_strerror (errno));
-          break;
-        }
-
-      textlen -= len;
-      pos += len;
-    } while (textlen > 0);
-
-  if (ok)
-    ftp_log (gftp_logging_misc, NULL,
-             _("Successfully wrote the log file to %s\n"), filename);
-
-  fclose (fd);
-  g_free (txt);
-}
-
-
 void 
 clear_cache (gpointer data)
 {
@@ -635,33 +436,44 @@ clear_cache (gpointer data)
 void 
 about_dialog (gpointer data)
 {
-  GtkWidget * tempwid, * notebook, * box, * label, * view, * vscroll,
-            * dialog;
-  char *tempstr, *no_license_agreement, *str, buf[255];
+  GtkWidget * tempwid, * notebook, * box, * label, * view, * vscroll, * dialog;
+  char *tempstr, *temp1str, *no_license_agreement, *str, buf[255], *share_dir;
   size_t len;
   FILE * fd;
-#if !(GTK_MAJOR_VERSION == 1 && GTK_MINOR_VERSION == 2)
+#if GTK_MAJOR_VERSION > 1
   GtkTextBuffer * textbuf;
   GtkTextIter iter;
-  guint textlen;
+  gint textlen;
 #endif
 
-  no_license_agreement = g_strdup_printf (_("Cannot find the license agreement file COPYING. Please make sure it is in either %s or in %s"), BASE_CONF_DIR, SHARE_DIR);
+  share_dir = gftp_get_share_dir ();
+  no_license_agreement = g_strdup_printf (_("Cannot find the license agreement file COPYING. Please make sure it is in either %s or in %s"), BASE_CONF_DIR, share_dir);
 
+#if GTK_MAJOR_VERSION == 1
   dialog = gtk_dialog_new ();
   gtk_window_set_title (GTK_WINDOW (dialog), _("About gFTP"));
-  gtk_window_set_wmclass (GTK_WINDOW(dialog), "about", "gFTP");
-  gtk_container_border_width (GTK_CONTAINER (GTK_DIALOG (dialog)->vbox), 10);
   gtk_container_border_width (GTK_CONTAINER
 			      (GTK_DIALOG (dialog)->action_area), 5);
-  gtk_box_set_spacing (GTK_BOX (GTK_DIALOG (dialog)->vbox), 5);
   gtk_box_set_homogeneous (GTK_BOX (GTK_DIALOG (dialog)->action_area), TRUE);
+#else
+  dialog = gtk_dialog_new_with_buttons (_("About gFTP"), NULL, 0,
+                                        GTK_STOCK_CLOSE,
+                                        GTK_RESPONSE_CLOSE,
+                                        NULL);
+#endif
+  gtk_window_set_wmclass (GTK_WINDOW(dialog), "about", "gFTP");
   gtk_window_set_position (GTK_WINDOW (dialog), GTK_WIN_POS_MOUSE);
-  gtk_signal_connect_object (GTK_OBJECT (dialog), "delete_event",
-			     GTK_SIGNAL_FUNC (gtk_widget_destroy),
-			     GTK_OBJECT (dialog));
+  gtk_container_border_width (GTK_CONTAINER (GTK_DIALOG (dialog)->vbox), 10);
+  gtk_box_set_spacing (GTK_BOX (GTK_DIALOG (dialog)->vbox), 5);
   gtk_widget_realize (dialog);
 
+  if (gftp_icon != NULL)
+    {
+      gdk_window_set_icon (dialog->window, NULL, gftp_icon->pixmap,
+                           gftp_icon->bitmap);
+      gdk_window_set_icon_name (dialog->window, gftp_version);
+    }
+
   notebook = gtk_notebook_new ();
   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), notebook, TRUE,
 		      TRUE, 0);
@@ -675,12 +487,12 @@ about_dialog (gpointer data)
   gtk_box_pack_start (GTK_BOX (box), tempwid, FALSE, FALSE, 0);
   gtk_widget_show (tempwid);
 
-  tempstr = g_strdup_printf (_("%s\nCopyright (C) 1998-2002 Brian Masney <masneyb@seul.org>\nOfficial Homepage: http://gftp.seul.org/\nLogo by: Aaron Worley <planet_hoth@yahoo.com>\n"),
-version);
+  tempstr = g_strdup_printf (_("%s\nCopyright (C) 1998-2007 Brian Masney <masneyb@gftp.org>\nOfficial Homepage: http://www.gftp.org/\n"), gftp_version);
   str = _("Translated by");
   if (strcmp (str, "Translated by") != 0)
     {
-      tempstr = g_realloc (tempstr, strlen (tempstr) + strlen (str) + 1);
+      tempstr = g_realloc (tempstr,
+                           (gulong) (strlen (tempstr) + strlen (str) + 1));
       strcat (tempstr, str);
     }
   tempwid = gtk_label_new (tempstr);
@@ -701,7 +513,7 @@ version);
   gtk_box_pack_start (GTK_BOX (box), tempwid, TRUE, TRUE, 0);
   gtk_widget_show (tempwid);
 
-#if GTK_MAJOR_VERSION == 1 && GTK_MINOR_VERSION == 2
+#if GTK_MAJOR_VERSION == 1
   view = gtk_text_new (NULL, NULL);
   gtk_text_set_editable (GTK_TEXT (view), FALSE);
   gtk_text_set_word_wrap (GTK_TEXT (view), TRUE);
@@ -718,6 +530,7 @@ version);
 #else
   view = gtk_text_view_new ();
   gtk_text_view_set_editable (GTK_TEXT_VIEW (view), FALSE);
+  gtk_text_view_set_cursor_visible (GTK_TEXT_VIEW (view), FALSE);
   gtk_text_view_set_wrap_mode (GTK_TEXT_VIEW (view), GTK_WRAP_WORD);
 
   vscroll = gtk_scrolled_window_new (NULL, NULL);
@@ -741,6 +554,7 @@ version);
 
   gtk_notebook_append_page (GTK_NOTEBOOK (notebook), box, label);
 
+#if GTK_MAJOR_VERSION == 1
   tempwid = gtk_button_new_with_label (_("  Close  "));
   GTK_WIDGET_SET_FLAGS (tempwid, GTK_CAN_DEFAULT);
   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area), tempwid,
@@ -750,24 +564,31 @@ version);
 			     GTK_OBJECT (dialog));
   gtk_widget_grab_default (tempwid);
   gtk_widget_show (tempwid);
+#else
+  g_signal_connect_swapped (GTK_OBJECT (dialog), "response",
+                            G_CALLBACK (gtk_widget_destroy),
+                            GTK_OBJECT (dialog));
+#endif
 
   tempstr = g_strconcat ("/usr/share/common-licenses/GPL", NULL);
   if (access (tempstr, F_OK) != 0)
     {
       g_free (tempstr);
-      tempstr = expand_path (SHARE_DIR "/COPYING");
+      temp1str = g_strconcat (share_dir, "/COPYING", NULL);
+      tempstr = gftp_expand_path (NULL, temp1str);
+      g_free (temp1str);
       if (access (tempstr, F_OK) != 0)
 	{
 	  g_free (tempstr);
-          tempstr = expand_path (BASE_CONF_DIR "/COPYING");
+          tempstr = gftp_expand_path (NULL, BASE_CONF_DIR "/COPYING");
 	  if (access (tempstr, F_OK) != 0)
 	    {
-#if GTK_MAJOR_VERSION == 1 && GTK_MINOR_VERSION == 2
+#if GTK_MAJOR_VERSION == 1
 	      gtk_text_insert (GTK_TEXT (view), NULL, NULL, NULL,
 			       no_license_agreement, -1);
 #else
               textlen = gtk_text_buffer_get_char_count (textbuf);
-              gtk_text_buffer_get_iter_at_offset (textbuf, &iter, textlen - 1);
+              gtk_text_buffer_get_iter_at_offset (textbuf, &iter, textlen);
               gtk_text_buffer_insert (textbuf, &iter, no_license_agreement, -1);
 #endif
 	      gtk_widget_show (dialog);
@@ -778,12 +599,12 @@ version);
 
   if ((fd = fopen (tempstr, "r")) == NULL)
     {
-#if GTK_MAJOR_VERSION == 1 && GTK_MINOR_VERSION == 2
+#if GTK_MAJOR_VERSION == 1
       gtk_text_insert (GTK_TEXT (view), NULL, NULL, NULL,
 		       no_license_agreement, -1);
 #else
       textlen = gtk_text_buffer_get_char_count (textbuf);
-      gtk_text_buffer_get_iter_at_offset (textbuf, &iter, textlen - 1);
+      gtk_text_buffer_get_iter_at_offset (textbuf, &iter, textlen);
       gtk_text_buffer_insert (textbuf, &iter, no_license_agreement, -1);
 #endif
       gtk_widget_show (dialog);
@@ -796,11 +617,11 @@ version);
   while ((len = fread (buf, 1, sizeof (buf) - 1, fd)))
     {
       buf[len] = '\0';
-#if GTK_MAJOR_VERSION == 1 && GTK_MINOR_VERSION == 2
+#if GTK_MAJOR_VERSION == 1
       gtk_text_insert (GTK_TEXT (view), NULL, NULL, NULL, buf, -1);
 #else
       textlen = gtk_text_buffer_get_char_count (textbuf);
-      gtk_text_buffer_get_iter_at_offset (textbuf, &iter, textlen - 1);
+      gtk_text_buffer_get_iter_at_offset (textbuf, &iter, textlen);
       gtk_text_buffer_insert (textbuf, &iter, buf, -1);
 #endif
     }
@@ -811,81 +632,62 @@ version);
 }
 
 
-void 
-compare_windows (gpointer data)
+static void
+_do_compare_windows (gftp_window_data * win1, gftp_window_data * win2)
 {
   gftp_file * curfle, * otherfle;
   GList * curlist, * otherlist;
-  int row;
+  int row, curdir, othdir;
 
+  row = 0;
+  curlist = win1->files;
+  while (curlist != NULL)
+    {
+      curfle = curlist->data;
+      if (!curfle->shown)
+        {
+          curlist = curlist->next;
+          continue;
+        }
+
+      otherlist = win2->files;
+      while (otherlist != NULL)
+	{
+          otherfle = otherlist->data;
+          if (!otherfle->shown)
+            {
+              otherlist = otherlist->next;
+              continue;
+            }
+
+          curdir = S_ISDIR (curfle->st_mode);
+          othdir = S_ISDIR (otherfle->st_mode);
+
+          if (strcmp (otherfle->file, curfle->file) == 0 &&
+              curdir == othdir &&
+              (curdir || otherfle->size == curfle->size))
+	    break;
+
+          otherlist = otherlist->next;
+	}
+
+      if (otherlist == NULL)
+	gtk_clist_select_row (GTK_CLIST (win1->listbox), row, 0);
+      row++;
+      curlist = curlist->next;
+    }
+}
+
+
+void 
+compare_windows (gpointer data)
+{
   if (!check_status (_("Compare Windows"), &window2, 1, 0, 0, 1))
     return;
 
   deselectall (&window1);
   deselectall (&window2);
 
-  row = 0;
-  curlist = window1.files;
-  while (curlist != NULL)
-    {
-      curfle = curlist->data;
-      if (!curfle->shown)
-        {
-          curlist = curlist->next;
-          continue;
-        }
-
-      otherlist = window2.files;
-      while (otherlist != NULL)
-	{
-          otherfle = otherlist->data;
-          if (!otherfle->shown)
-            {
-              otherlist = otherlist->next;
-              continue;
-            }
-	  if (strcmp (otherfle->file, curfle->file) == 0
-	      && otherfle->size == curfle->size)
-	    break;
-          otherlist = otherlist->next;
-	}
-
-      if (otherlist == NULL)
-	gtk_clist_select_row (GTK_CLIST (window1.listbox), row, 0);
-      row++;
-      curlist = curlist->next;
-    }
-
-  row = 0;
-  curlist = window2.files;
-  while (curlist != NULL)
-    {
-      curfle = curlist->data;
-      if (!curfle->shown)
-        {
-          curlist = curlist->next;
-          continue;
-        }
-
-      otherlist = window1.files;
-      while (otherlist != NULL)
-	{
-          otherfle = otherlist->data;
-          if (!otherfle->shown)
-            {
-              otherlist = otherlist->next;
-              continue;
-            }
-	  if (strcmp (otherfle->file, curfle->file) == 0
-	      && otherfle->size == curfle->size)
-	    break;
-          otherlist = otherlist->next;
-	}
-
-      if (otherlist == NULL)
-	gtk_clist_select_row (GTK_CLIST (window2.listbox), row, 0);
-      row++;
-      curlist = curlist->next;
-    }
+  _do_compare_windows (&window1, &window2);
+  _do_compare_windows (&window2, &window1);
 }
-
