@@ -1,6 +1,6 @@
 /*****************************************************************************/
 /*  gftpui.c - UI related functions for gFTP                                 */
-/*  Copyright (C) 1998-2003 Brian Masney <masneyb@gftp.org>                  */
+/*  Copyright (C) 1998-2007 Brian Masney <masneyb@gftp.org>                  */
 /*                                                                           */
 /*  This program is free software; you can redistribute it and/or modify     */
 /*  it under the terms of the GNU General Public License as published by     */
@@ -18,91 +18,60 @@
 /*****************************************************************************/
 
 #include "gftpui.h"
-static const char cvsid[] = "$Id: gftpui.c,v 1.36 2005/01/25 02:34:19 masneyb Exp $";
-
-sigjmp_buf gftpui_common_jmp_environment;
-volatile int gftpui_common_use_jmp_environment = 0;
+static const char cvsid[] = "$Id: gftpui.c 980 2008-11-30 20:03:50Z masneyb $";
 
 GStaticMutex gftpui_common_transfer_mutex = G_STATIC_MUTEX_INIT;
 volatile sig_atomic_t gftpui_common_child_process_done = 0;
+volatile sig_atomic_t gftpui_common_num_child_threads = 0;
 static gftp_logging_func gftpui_common_logfunc;
 
-
-static int
-_gftpui_cb_connect (gftpui_callback_data * cdata)
-{
-  if (cdata->connect_function != NULL)
-    return (cdata->connect_function (cdata));
-  else
-    return (gftp_connect (cdata->request));
-}
-
-
-static void
-_gftpui_cb_disconnect (gftpui_callback_data * cdata)
-{
-  if (cdata->connect_function != NULL)
-    cdata->disconnect_function (cdata);
-  else
-    gftp_disconnect (cdata->request);
-}
-
-
 static void *
 _gftpui_common_thread_callback (void * data)
 { 
   intptr_t network_timeout, sleep_time;
   gftpui_callback_data * cdata;
-  int success, sj, num_timeouts;
+  struct timespec ts;
+  int success;
 
   cdata = data;
+  gftpui_common_num_child_threads++;
+
   gftp_lookup_request_option (cdata->request, "network_timeout",
                               &network_timeout);
   gftp_lookup_request_option (cdata->request, "sleep_time",
                               &sleep_time);
 
-  sj = sigsetjmp (gftpui_common_jmp_environment, 1);
-  gftpui_common_use_jmp_environment = 1;
-
-  num_timeouts = 0;
   success = GFTP_ERETRYABLE;
-  if (sj != 1)
+  while (1)
     {
-      while (1)
+      if (network_timeout > 0)
+        alarm (network_timeout);
+
+      success = cdata->run_function (cdata);
+      alarm (0);
+
+      if (cdata->request->cancel)
         {
-          if (network_timeout > 0)
-            alarm (network_timeout);
-          success = cdata->run_function (cdata);
-          alarm (0);
-
-          if (success == GFTP_ETIMEDOUT && num_timeouts == 0)
-            {
-              _gftpui_cb_disconnect (cdata);
-              num_timeouts++;
-              if (_gftpui_cb_connect (cdata) == 0)
-                continue;
-            }
-
-          if (success == GFTP_EFATAL || success == 0 || cdata->retries == 0)
-            break;
-
-          cdata->retries--;
           cdata->request->logging_function (gftp_logging_error, cdata->request,
-                       _("Waiting %d seconds until trying to connect again\n"),
-                       sleep_time);
-          alarm (sleep_time);
-          pause ();
+                                            _("Operation canceled\n"));
+          break;
         }
-    }
-  else
-    {
-      _gftpui_cb_disconnect (cdata);
+        
+      if (success == GFTP_EFATAL || success == 0 || cdata->retries == 0)
+        break;
+
+      cdata->retries--;
       cdata->request->logging_function (gftp_logging_error, cdata->request,
-                                        _("Operation canceled\n"));
+                   _("Waiting %d seconds until trying to connect again\n"),
+                   sleep_time);
+
+      ts.tv_sec = sleep_time;
+      ts.tv_nsec = 0;
+      nanosleep (&ts, NULL);
     }
 
-  gftpui_common_use_jmp_environment = 0;
   cdata->request->stopable = 0;
+  gftpui_common_num_child_threads--;
 
   return (GINT_TO_POINTER (success));
 }
@@ -133,10 +102,8 @@ gftpui_common_signal_handler (int signo)
 {
   signal (signo, gftpui_common_signal_handler);
 
-  if (gftpui_common_use_jmp_environment)
-    siglongjmp (gftpui_common_jmp_environment, signo == SIGINT ? 1 : 2);
-  else if (signo == SIGINT)
-    exit (1);
+  if (!gftpui_common_num_child_threads && signo == SIGINT)
+    exit (EXIT_FAILURE);
 }
 
 
@@ -173,7 +140,7 @@ gftpui_common_init (int *argc, char ***argv, gftp_logging_func logfunc)
   share_dir = gftp_get_share_dir ();
   gftp_read_config_file (share_dir);
   if (gftp_parse_command_line (argc, argv) != 0)
-    exit (0);
+    exit (EXIT_FAILURE);
 
   gftpui_common_logfunc = logfunc;
   gftpui_common_child_process_done = -1;
@@ -185,9 +152,7 @@ gftpui_common_about (gftp_logging_func logging_function, gpointer logdata)
 {
   char *str;
 
-  logging_function (gftp_logging_misc, logdata, "%s, Copyright (C) 1998-2003 Brian Masney <", gftp_version);
-  logging_function (gftp_logging_recv, logdata, "masneyb@gftp.org");
-  logging_function (gftp_logging_misc, logdata, _(">. If you have any questions, comments, or suggestions about this program, please feel free to email them to me. You can always find out the latest news about gFTP from my website at http://www.gftp.org/\n"));
+  logging_function (gftp_logging_misc, logdata, "%s, Copyright (C) 1998-2008 Brian Masney <masneyb@gftp.org>. If you have any questions, comments, or suggestions about this program, please feel free to email them to me. You can always find out the latest news about gFTP from my website at http://www.gftp.org/\n", gftp_version);
   logging_function (gftp_logging_misc, logdata, _("gFTP comes with ABSOLUTELY NO WARRANTY; for details, see the COPYING file. This is free software, and you are welcome to redistribute it under certain conditions; for details, see the COPYING file\n"));
 
   str = _("Translated by");
@@ -404,6 +369,7 @@ gftpui_common_cmd_site (void *uidata, gftp_request * request,
       cdata->uidata = uidata;
       cdata->input_string = (char *) command;
       cdata->run_function = gftpui_common_run_site;
+      cdata->toggled = 1;
 
       gftpui_common_run_callback_function (cdata);
 
@@ -625,7 +591,7 @@ gftpui_common_cmd_ls (void *uidata, gftp_request * request,
       fle = templist->data;
 
       gftpui_lookup_file_colors (fle, &startcolor, &endcolor);
-      tempstr = gftp_gen_ls_string (fle, startcolor, endcolor);
+      tempstr = gftp_gen_ls_string (request, fle, startcolor, endcolor);
       request->logging_function (gftp_logging_misc_nolog, request, "%s\n",
                                  tempstr);
       g_free (tempstr);
@@ -682,7 +648,12 @@ gftpui_common_cmd_open (void *uidata, gftp_request * request,
   cdata->retries = retries;
   cdata->dont_check_connection = 1;
 
+  if (request->refreshing)
+    cdata->dont_refresh = 1;
+
+  gftpui_show_busy (TRUE);
   gftpui_common_run_callback_function (cdata);
+  gftpui_show_busy (FALSE);
 
   g_free (cdata);
 
@@ -838,6 +809,7 @@ gftpui_common_cmd_help (void *uidata, gftp_request * request,
       for (i=0; i<numrows; i++)
         {
           strncpy (commands, "\t", sizeof (commands));
+          size_t cmd_len = sizeof(commands) - strlen (commands);
 
           for (j=0; j<numcols; j++)
             {
@@ -847,7 +819,8 @@ gftpui_common_cmd_help (void *uidata, gftp_request * request,
 
               g_snprintf (cmdstr, sizeof (cmdstr), "%-10s",
                           gftpui_common_commands[ele].command);
-              strncat (commands, cmdstr, sizeof (commands));
+              strncat (commands, cmdstr, cmd_len);
+              cmd_len -= strlen(cmdstr);
             }
           gftpui_common_logfunc (gftp_logging_misc_nolog, request, "%s\n",
                                  commands);
@@ -858,9 +831,9 @@ gftpui_common_cmd_help (void *uidata, gftp_request * request,
 
 
 static void
-_gftpui_common_transfer_files (void *fromuidata, gftp_request * fromrequest,
-                               void *touidata, gftp_request * torequest,
-                               const char *cmd, const char *filespec)
+_gftpui_common_cmd_transfer_files (void *fromuidata, gftp_request * fromrequest,
+                                   void *touidata, gftp_request * torequest,
+                                   const char *cmd, const char *filespec)
 {
   gftp_transfer * tdata;
   gftp_file * fle;
@@ -901,7 +874,7 @@ _gftpui_common_transfer_files (void *fromuidata, gftp_request * fromrequest,
         }
 
       tdata->files = g_list_append (tdata->files, fle);
-      fle = g_malloc (sizeof (*fle));
+      fle = g_malloc0 (sizeof (*fle));
     }
 
   g_free (fle);
@@ -943,8 +916,8 @@ gftpui_common_cmd_mget_file (void *uidata, gftp_request * request,
                              void *other_uidata, gftp_request * other_request,
                              const char *command)
 {
-  _gftpui_common_transfer_files (uidata, request, other_uidata, other_request,
-                                 "mget", command);
+  _gftpui_common_cmd_transfer_files (uidata, request, other_uidata,
+                                     other_request, "mget", command);
   return (1);
 }
 
@@ -954,8 +927,8 @@ gftpui_common_cmd_mput_file (void *uidata, gftp_request * request,
                              void *other_uidata, gftp_request * other_request,
                              const char *command)
 {
-  _gftpui_common_transfer_files (other_uidata, other_request, uidata, request,
-                                 "mput", command);
+  _gftpui_common_cmd_transfer_files (other_uidata, other_request, uidata,
+                                     request, "mput", command);
   return (1);
 }
 
@@ -1058,11 +1031,10 @@ gftpui_common_process_command (void *locui, gftp_request * locreq,
     newstr[--len] = '\0';
 
   for (pos = newstr + len - 1;
-       (*pos == ' ' || *pos == '\t') && pos > newstr;
-       pos--)
-    *pos = '\0';
+       (*pos == ' ' || *pos == '\t') && pos > newstr; 
+       *pos-- = '\0');
 
-  if (*stpos == '\0')
+  if (*newstr == '\0')
     {
       g_free (newstr);
       return (1);
@@ -1133,24 +1105,29 @@ gftpui_common_add_file_transfer (gftp_request * fromreq, gftp_request * toreq,
                                  void *fromuidata, void *touidata,
                                  GList * files)
 {
-  intptr_t append_transfers, one_transfer;
+  intptr_t append_transfers, one_transfer, overwrite_default;
   GList * templist, *curfle;
   gftp_transfer * tdata;
   gftp_file * tempfle;
   int show_dialog;
   
-  for (templist = files; templist != NULL; templist = templist->next)
-    { 
-      tempfle = templist->data;
-      if (tempfle->startsize > 0)
-        break;
+  gftp_lookup_request_option (fromreq, "overwrite_default", &overwrite_default);
+  gftp_lookup_request_option (fromreq, "append_transfers", &append_transfers);
+  gftp_lookup_request_option (fromreq, "one_transfer", &one_transfer);
+
+  if (!overwrite_default)
+    {
+      for (templist = files; templist != NULL; templist = templist->next)
+        { 
+          tempfle = templist->data;
+          if (tempfle->startsize > 0 && !S_ISDIR (tempfle->st_mode))
+            break;
+        }
+
+      show_dialog = templist != NULL;
     }
-  show_dialog = templist != NULL;
-  
-  gftp_lookup_request_option (fromreq, "append_transfers",
-                              &append_transfers);
-  gftp_lookup_request_option (fromreq, "one_transfer",
-                              &one_transfer);
+  else
+    show_dialog = 0;
 
   tdata = NULL;
   if (append_transfers && one_transfer && !show_dialog)
@@ -1248,145 +1225,261 @@ gftpui_common_add_file_transfer (gftp_request * fromreq, gftp_request * toreq,
 }
 
 
-static void
-_gftpui_common_setup_fds (gftp_transfer * tdata, gftp_file * curfle,
-                          int *fromfd, int *tofd)
+static ssize_t
+_do_transfer_block (gftp_transfer * tdata, gftp_file * curfle, char *buf,
+                    size_t trans_blksize)
 {
-  *tofd = -1;
-  *fromfd = -1;
+  ssize_t num_read, num_wrote, ret;
+  char *bufpos;
 
-  if (curfle->is_fd)
+  num_read = gftp_get_next_file_chunk (tdata->fromreq, buf, trans_blksize);
+  if (num_read < 0)
+    return (num_read);
+
+  bufpos = buf;
+  num_wrote = 0;
+  while (num_wrote < num_read)
     {
-      if (tdata->toreq->protonum == GFTP_LOCAL_NUM)
-        *tofd = curfle->fd;
-      else if (tdata->fromreq->protonum == GFTP_LOCAL_NUM)
-        *fromfd = curfle->fd;
-    }
-}
+      if ((ret = gftp_put_next_file_chunk (tdata->toreq, bufpos,
+                                           num_read - num_wrote)) <= 0)
+        return (ret);
 
-
-static void
-_gftpui_common_done_with_fds (gftp_transfer * tdata, gftp_file * curfle)
-{
-  if (curfle->is_fd)
-    {
-      if (tdata->toreq->protonum == GFTP_LOCAL_NUM)
-        tdata->toreq->datafd = -1;
-      else
-        tdata->fromreq->datafd = -1;
+      num_wrote += ret;
+      bufpos += ret;
     }
+
+  return (num_read);
 }
 
 
 int
-gftpui_common_transfer_files (gftp_transfer * tdata)
+_gftpui_common_do_transfer_file (gftp_transfer * tdata, gftp_file * curfle)
 {
-  intptr_t preserve_permissions, preserve_time, trans_blksize;
   struct timeval updatetime;
-  ssize_t num_read, ret;
-  gftp_file * curfle; 
-  int tofd, fromfd;
+  intptr_t trans_blksize;
+  ssize_t num_trans;
   char *buf;
+  int ret;
 
   gftp_lookup_request_option (tdata->fromreq, "trans_blksize", &trans_blksize);
-  buf = g_malloc (trans_blksize);
+  buf = g_malloc0 (trans_blksize);
 
-  tdata->curfle = tdata->files;
-  gettimeofday (&tdata->starttime, NULL);
-  memcpy (&tdata->lasttime, &tdata->starttime, sizeof (tdata->lasttime));
+  memset (&updatetime, 0, sizeof (updatetime));
+  gftpui_start_current_file_in_transfer (tdata);
 
-  while (tdata->curfle != NULL)
+  num_trans = 0;
+  while (!tdata->cancel &&
+         (num_trans = _do_transfer_block (tdata, curfle, buf,
+                                          trans_blksize)) > 0)
     {
-      num_read = -1;
+      gftp_calc_kbs (tdata, num_trans);
 
-      if (g_thread_supported ())
-        g_static_mutex_lock (&tdata->structmutex);
-
-      curfle = tdata->curfle->data;
-      tdata->current_file_number++;
-
-      if (g_thread_supported ())
-        g_static_mutex_unlock (&tdata->structmutex);
-
-      if (curfle->transfer_action == GFTP_TRANS_ACTION_SKIP)
+      if (tdata->lasttime.tv_sec - updatetime.tv_sec >= 1 ||
+          tdata->curtrans >= tdata->tot_file_trans)
         {
-          if (g_thread_supported ())
-            g_static_mutex_lock (&tdata->structmutex);
+          gftpui_update_current_file_in_transfer (tdata);
+          memcpy (&updatetime, &tdata->lasttime, sizeof (updatetime));
 
-          tdata->next_file = 1;
-          tdata->curfle = tdata->curfle->next;
+          if (tdata->current_file_retries > 0)
+            tdata->current_file_retries = 0;
+        }
+    }
 
-          if (g_thread_supported ())
-            g_static_mutex_unlock (&tdata->structmutex);
+  if (num_trans == GFTP_ENOTRANS)
+    num_trans = 0;
 
-          continue;
+  g_free (buf);
+  gftpui_finish_current_file_in_transfer (tdata);
+
+  if ((int) num_trans == 0)
+    {
+      if ((ret = gftp_end_transfer (tdata->fromreq)) < 0)
+        return (ret);
+
+      if ((ret = gftp_end_transfer (tdata->toreq)) < 0)
+        return (ret);
+
+      tdata->fromreq->logging_function (gftp_logging_misc,
+                     tdata->fromreq,
+                     _("Successfully transferred %s at %.2f KB/s\n"),
+                     curfle->file, tdata->kbs);
+
+      return (0);
+    }
+  else
+    return ((int) num_trans);
+}
+
+
+void
+gftpui_common_skip_file_transfer (gftp_transfer * tdata, gftp_file * curfle)
+{
+  g_static_mutex_lock (&tdata->structmutex);
+
+  if (tdata->started && !(curfle->transfer_action & GFTP_TRANS_ACTION_SKIP))
+    {
+      curfle->transfer_action = GFTP_TRANS_ACTION_SKIP;
+      if (tdata->curfle != NULL && curfle == tdata->curfle->data)
+        {
+          gftpui_cancel_file_transfer (tdata);
+          tdata->skip_file = 1;
+        }
+      else if (!curfle->transfer_done)
+        tdata->total_bytes -= curfle->size;
+    }
+
+  g_static_mutex_unlock (&tdata->structmutex);
+
+  if (curfle != NULL)
+    tdata->fromreq->logging_function (gftp_logging_misc, tdata->fromreq,
+                                      _("Skipping file %s on host %s\n"),
+                                      curfle->file, tdata->toreq->hostname);
+}
+
+
+void
+gftpui_common_cancel_file_transfer (gftp_transfer * tdata)
+{
+  g_static_mutex_lock (&tdata->structmutex);
+
+  if (tdata->started)
+    {
+      gftpui_cancel_file_transfer (tdata);
+      tdata->skip_file = 0;
+    }
+  else
+    tdata->done = 1;
+
+  tdata->fromreq->stopable = 0;
+  tdata->toreq->stopable = 0;
+
+  g_static_mutex_unlock (&tdata->structmutex);
+
+  tdata->fromreq->logging_function (gftp_logging_misc, tdata->fromreq,
+                                    _("Stopping the transfer on host %s\n"),
+                                    tdata->toreq->hostname);
+}
+
+
+static void
+_gftpui_common_next_file_in_trans (gftp_transfer * tdata)
+{
+  gftp_file * curfle;
+
+  if (g_thread_supported ())
+    g_static_mutex_lock (&tdata->structmutex);
+
+  tdata->curtrans = 0;
+  tdata->next_file = 1;
+
+  curfle = tdata->curfle->data;
+  curfle->transfer_done = 1;
+  tdata->curfle = tdata->curfle->next;
+
+  if (g_thread_supported ())
+    g_static_mutex_unlock (&tdata->structmutex);
+}
+
+
+static int
+_gftpui_common_preserve_perm_time (gftp_transfer * tdata, gftp_file * curfle)
+{
+  intptr_t preserve_permissions, preserve_time;
+  int ret, tmpret;
+
+  gftp_lookup_request_option (tdata->fromreq, "preserve_permissions",
+                              &preserve_permissions);
+  gftp_lookup_request_option (tdata->fromreq, "preserve_time",
+                              &preserve_time);
+
+  ret = 0;
+  if (GFTP_IS_CONNECTED (tdata->toreq) && preserve_permissions &&
+      curfle->st_mode != 0)
+    {
+      tmpret = gftp_chmod (tdata->toreq, curfle->destfile,
+                           curfle->st_mode & (S_IRWXU | S_IRWXG | S_IRWXO));
+      if (tmpret < 0)
+        ret = tmpret;
+    }
+
+  if (GFTP_IS_CONNECTED (tdata->toreq) && preserve_time &&
+      curfle->datetime != 0)
+    {
+      tmpret = gftp_set_file_time (tdata->toreq, curfle->destfile,
+                                   curfle->datetime);
+      if (tmpret < 0)
+        ret = tmpret;
+    }
+
+  if (!GFTP_IS_CONNECTED (tdata->toreq))
+    return (ret);
+  else
+    return (0);
+}
+
+
+static int
+_gftpui_common_trans_file_or_dir (gftp_transfer * tdata)
+{
+  gftp_file * curfle;
+  int ret;
+
+  if (g_thread_supported ())
+    g_static_mutex_lock (&tdata->structmutex);
+
+  curfle = tdata->curfle->data;
+  tdata->current_file_number++;
+
+  if (g_thread_supported ())
+    g_static_mutex_unlock (&tdata->structmutex);
+
+  if (curfle->transfer_action == GFTP_TRANS_ACTION_SKIP)
+    {
+      tdata->tot_file_trans = 0;
+      return (0);
+    }
+
+  if ((ret = gftp_connect (tdata->fromreq)) < 0)
+    return (ret);
+
+  if ((ret = gftp_connect (tdata->toreq)) < 0)
+    return (ret);
+
+  if (S_ISDIR (curfle->st_mode))
+    {
+      tdata->tot_file_trans = 0;
+      if (curfle->startsize > 0)
+        ret = 1;
+      else
+        ret = gftp_make_directory (tdata->toreq, curfle->destfile);
+    }
+  else
+    {
+      if (curfle->size == 0)
+        {
+          curfle->size = gftp_get_file_size (tdata->fromreq, curfle->file);
+          if (curfle->size < 0)
+            return ((int) curfle->size);
+
+          tdata->total_bytes += curfle->size;
         }
 
-      tdata->tot_file_trans = -1;
-      if (gftp_connect (tdata->fromreq) == 0 &&
-          gftp_connect (tdata->toreq) == 0)
+      if (curfle->retry_transfer)
         {
-          if (S_ISDIR (curfle->st_mode))
-            {
-              if (tdata->toreq->mkdir != NULL)
-                {
-                  tdata->toreq->mkdir (tdata->toreq, curfle->destfile);
-                  if (!GFTP_IS_CONNECTED (tdata->toreq))
-                    break;
-                }
-
-              if (g_thread_supported ())
-                g_static_mutex_lock (&tdata->structmutex);
-
-              tdata->next_file = 1;
-              tdata->curfle = tdata->curfle->next;
-
-              if (g_thread_supported ())
-                g_static_mutex_unlock (&tdata->structmutex);
-              continue;
-            }
-
-          _gftpui_common_setup_fds (tdata, curfle, &fromfd, &tofd);
-
-          if (curfle->size == 0)
-            {
-              curfle->size = gftp_get_file_size (tdata->fromreq, curfle->file);
-              tdata->total_bytes += curfle->size;
-            }
-
-          if (GFTP_IS_CONNECTED (tdata->fromreq) &&
-              GFTP_IS_CONNECTED (tdata->toreq))
-            {
-              tdata->tot_file_trans = gftp_transfer_file (tdata->fromreq,
-                          curfle->file, fromfd,
-                          curfle->transfer_action == GFTP_TRANS_ACTION_RESUME ?
-                                                    curfle->startsize : 0,
-                          tdata->toreq, curfle->destfile, tofd,
-                          curfle->transfer_action == GFTP_TRANS_ACTION_RESUME ?
-                                                    curfle->startsize : 0);
-            }
+          curfle->transfer_action = GFTP_TRANS_ACTION_RESUME;
+          curfle->startsize = gftp_get_file_size (tdata->toreq, curfle->destfile);
+          if (curfle->startsize < 0)
+            return ((int) curfle->startsize);
         }
 
-      if (!GFTP_IS_CONNECTED (tdata->fromreq) ||
-          !GFTP_IS_CONNECTED (tdata->toreq))
-        {
-          tdata->fromreq->logging_function (gftp_logging_error,
-                         tdata->fromreq,
-                         _("Error: Remote site disconnected after trying to transfer file\n"));
-        }
-      else if (tdata->tot_file_trans < 0)
-        {
-          if (g_thread_supported ())
-            g_static_mutex_lock (&tdata->structmutex);
-
-          curfle->transfer_action = GFTP_TRANS_ACTION_SKIP;
-          tdata->next_file = 1;
-          tdata->curfle = tdata->curfle->next;
-
-          if (g_thread_supported ())
-          g_static_mutex_unlock (&tdata->structmutex);
-          continue;
-        }
+      tdata->tot_file_trans = gftp_transfer_file (tdata->fromreq, curfle->file,
+                                                  curfle->transfer_action == GFTP_TRANS_ACTION_RESUME ?
+                                                          curfle->startsize : 0,
+                                                  tdata->toreq, curfle->destfile,
+                                                  curfle->transfer_action == GFTP_TRANS_ACTION_RESUME ?
+                                                          curfle->startsize : 0);
+      if (tdata->tot_file_trans < 0)
+        ret = tdata->tot_file_trans;
       else
         {
           if (g_thread_supported ())
@@ -1399,119 +1492,77 @@ gftpui_common_transfer_files (gftp_transfer * tdata)
           if (g_thread_supported ())
             g_static_mutex_unlock (&tdata->structmutex);
 
-          memset (&updatetime, 0, sizeof (updatetime));
-          gftpui_start_current_file_in_transfer (tdata);
-
-          while (!tdata->cancel &&
-                 (num_read = gftp_get_next_file_chunk (tdata->fromreq,
-                                                       buf, trans_blksize)) > 0)
-            {
-              gftp_calc_kbs (tdata, num_read);
-              if (tdata->lasttime.tv_sec - updatetime.tv_sec >= 1 ||
-                  tdata->curtrans >= tdata->tot_file_trans)
-                {
-                  gftpui_update_current_file_in_transfer (tdata);
-                  memcpy (&updatetime, &tdata->lasttime, sizeof (updatetime));
-                }
-
-              if ((ret = gftp_put_next_file_chunk (tdata->toreq, buf,
-                                                   num_read)) < 0)
-                {
-                  num_read = (int) ret;
-                  break;
-                }
-            }
-
-          if (num_read == GFTP_ENOTRANS)
-            num_read = 0;
-
-          gftpui_finish_current_file_in_transfer (tdata);
+          ret = _gftpui_common_do_transfer_file (tdata, curfle);
         }
+    }
 
+  if (ret == 0)
+    ret = _gftpui_common_preserve_perm_time (tdata, curfle);
+  else
+    {
+      curfle->retry_transfer = 1;
+      tdata->fromreq->logging_function (gftp_logging_error, tdata->fromreq,
+                                        _("Could not download %s from %s\n"),
+                                        curfle->file, tdata->fromreq->hostname);
+    }
+
+  return (ret);
+}
+
+
+int
+gftpui_common_transfer_files (gftp_transfer * tdata)
+{
+  int ret, skipped_files;
+
+  tdata->curfle = tdata->files;
+  gftpui_common_num_child_threads++;
+
+  gettimeofday (&tdata->starttime, NULL);
+  memcpy (&tdata->lasttime, &tdata->starttime, sizeof (tdata->lasttime));
+
+  skipped_files = 0;
+  while (tdata->curfle != NULL)
+    {
+      ret = _gftpui_common_trans_file_or_dir (tdata);
       if (tdata->cancel)
         {
-          if (gftp_abort_transfer (tdata->fromreq) != 0)
-            gftp_disconnect (tdata->fromreq);
-
           if (gftp_abort_transfer (tdata->toreq) != 0)
             gftp_disconnect (tdata->toreq);
-        }
-      else if (num_read < 0)
-        {
-          tdata->fromreq->logging_function (gftp_logging_error,
-                                        tdata->fromreq,
-                                        _("Could not download %s from %s\n"),
-                                        curfle->file,
-                                        tdata->fromreq->hostname);
 
-          if (gftp_get_transfer_status (tdata, num_read) == GFTP_ERETRYABLE)
+          if (gftp_abort_transfer (tdata->fromreq) != 0)
+            gftp_disconnect (tdata->fromreq);
+        }
+      else if (ret == GFTP_EFATAL)
+        skipped_files++;
+      else if (ret < 0)
+        {
+          if (gftp_get_transfer_status (tdata, ret) == GFTP_ERETRYABLE)
             continue;
 
           break;
         }
-      else
+
+      _gftpui_common_next_file_in_trans (tdata);
+
+      if (tdata->cancel)
         {
-          _gftpui_common_done_with_fds (tdata, curfle);
-          if (gftp_end_transfer (tdata->fromreq) != 0)
-            {
-              if (gftp_get_transfer_status (tdata, -1) == GFTP_ERETRYABLE)
-                continue;
+          if (!tdata->skip_file)
+            break;
 
-              break;
-            }
-
-          if (gftp_end_transfer (tdata->toreq) == 0)
-            {
-              tdata->fromreq->logging_function (gftp_logging_misc,
-                             tdata->fromreq,
-                             _("Successfully transferred %s at %.2f KB/s\n"),
-                             curfle->file, tdata->kbs);
-            }
-          else
-            {
-              tdata->fromreq->logging_function (gftp_logging_error,
-                             tdata->fromreq,
-                             _("There was an error transfering the file %s"),
-                             curfle->file);
-            }
+          tdata->cancel = 0;
+          tdata->fromreq->cancel = 0;
+          tdata->toreq->cancel = 0;
         }
-
-      gftp_lookup_request_option (tdata->fromreq, "preserve_permissions",
-                                  &preserve_permissions);
-      gftp_lookup_request_option (tdata->fromreq, "preserve_time",
-                                  &preserve_time);
-
-      if (!curfle->is_fd)
-        {
-          if (preserve_permissions && curfle->st_mode != 0)
-            gftp_chmod (tdata->toreq, curfle->destfile,
-                        curfle->st_mode & (S_IRWXU | S_IRWXG | S_IRWXO));
-
-          if (preserve_time && curfle->datetime != 0)
-            gftp_set_file_time (tdata->toreq, curfle->destfile,
-                                curfle->datetime);
-        }
-
-      if (g_thread_supported ())
-        g_static_mutex_lock (&tdata->structmutex);
-
-      tdata->curtrans = 0;
-      tdata->next_file = 1;
-      curfle->transfer_done = 1;
-      tdata->curfle = tdata->curfle->next;
-
-      if (g_thread_supported ())
-        g_static_mutex_unlock (&tdata->structmutex);
-
-      if (tdata->cancel && !tdata->skip_file)
-        break;
-      tdata->cancel = 0;
-      tdata->fromreq->cancel = 0;
-      tdata->toreq->cancel = 0;
     }
 
+  if (skipped_files)
+    tdata->fromreq->logging_function (gftp_logging_error, tdata->fromreq,
+                                      _("There were %d files or directories that could not be transferred. Check the log for which items were not properly transferred."),
+                                      skipped_files);
+
   tdata->done = 1;
-  g_free (buf);
+  gftpui_common_num_child_threads--;
 
   return (1);
 }
